# -*- coding: utf-8 -*- #
# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Routes to/from Compute Engine VMs."""



import argparse
import io
import os
import re

from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.api_lib.compute import lister
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.compute import ssh_utils
from googlecloudsdk.command_lib.compute.diagnose import external_helper
from googlecloudsdk.command_lib.compute.diagnose import internal_helpers
from googlecloudsdk.command_lib.util.ssh import ssh
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core.console import console_io
from googlecloudsdk.core.util import files
import six


DETAILED_HELP = {
    'EXAMPLES':
        """\
        To route to/from Compute Engine virtual machine instances, run:

          $ {command}
        """,
}


class Routes(base_classes.BaseCommand):
  """Routes to/from Compute Engine virtual machine instances.

  Routes to/from Compute Engine virtual machine instances.

  NOTE: The name filtering will cycle through all the VMs in the project.
  Depending on the size of the project, this could be a considerable amount
  of work.

  If that is the case, use the --regexp flag to filter down the amount
  of VMs considered in the filtering.
  """

  detailed_help = DETAILED_HELP

  @classmethod
  def Args(cls, parser):
    _RoutesArgs.Args(parser)

  def Run(self, args):
    """Default run method implementation."""
    super(Routes, self).Run(args)

    self._use_accounts_service = False

    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    resource_registry = holder.resources
    ssh_helper = ssh_utils.BaseSSHCLIHelper()
    ssh_helper.Run(args)

    # We store always needed commands non-changing fields
    self._args = args
    self._ssh_helper = ssh_helper

    # We obtain generic parameters of the call
    project = properties.VALUES.core.project.GetOrFail()
    filters = _RoutesArgs.GetFilters(args)
    instances = _RoutesQueries.ObtainInstances(
        args.names,
        service=self.compute.instances,
        project=project,
        zones=args.zones,
        filters=filters,
        http=self.http,
        batch_url=self.batch_url)

    user = args.user
    if not user:
      user = ssh.GetDefaultSshUsername()

    # We unpack the flags
    dry_run = args.dry_run
    reverse_traceroute = args.reverse_traceroute
    traceroute_args = args.traceroute_args
    external_route_ip = args.external_route_ip

    internal_helpers.PrintHeader(instances)
    prompt = 'The following VMs will be tracerouted.'
    if instances and not dry_run and not console_io.PromptContinue(prompt):
      return
    # Sometimes the prompt would appear after the instance data
    log.out.flush()

    for instance in instances:
      header = 'Checking instance %s' % instance.name
      log.out.Print(header)
      log.out.Print('-' * len(header))

      try:
        self.TracerouteInstance(instance, traceroute_args, dry_run,
                                resource_registry)
      except exceptions.ToolException as e:
        log.error('Error routing to instance')
        log.error(six.text_type(e))
        continue

      if reverse_traceroute:
        try:
          has_traceroute = self.CheckTraceroute(instance, user, dry_run,
                                                resource_registry)
          if has_traceroute:
            # We obtain the self ip
            if not external_route_ip:
              external_route_ip = self.ObtainSelfIp(instance, user, dry_run,
                                                    resource_registry)
            if external_route_ip:
              self.ReverseTracerouteInstance(instance, user, external_route_ip,
                                             traceroute_args, dry_run,
                                             resource_registry)
            else:
              log.out.Print('Unable to obtain self ip. Aborting.')
          else:
            log.out.Print(
                'Please make sure traceroute is installed in PATH to move on.')
        except ssh.CommandError as e:
          log.error(six.text_type(e))
      log.out.Print('')  # Separator

  ###########################################################
  # Traceroute Invocations
  ###########################################################

  def TracerouteInstance(self, instance, traceroute_args, dry_run,
                         resource_registry):
    """Runs a traceroute from localhost to a GCE VM.

    Args:
      instance: Compute Engine VM.
      traceroute_args: Additional traceroute args to be passed on.
      dry_run: Whether to only print commands instead of running them.
      resource_registry: gcloud class used for obtaining data from the
        resources.
    """
    instance_string = internal_helpers.GetInstanceNetworkTitleString(instance)
    log.out.Print('>>> Tracerouting to %s' % instance_string)

    external_ip = ssh_utils.GetExternalIPAddress(instance)
    cmd = ['traceroute', external_ip]
    if traceroute_args:
      cmd += traceroute_args
    if dry_run:
      external_helper.DryRunLog(' '.join(cmd))
    else:
      external_helper.RunSubprocess(proc_name='Traceroute', command_list=cmd)
      log.out.Print('>>>')

  def ReverseTracerouteInstance(self, instance, user, external_route_ip,
                                traceroute_args, dry_run, resource_registry):
    """Runs a traceroute from a GCE VM to localhost.

    Args:
      instance: Compute Engine VM.
      user: The user to use to SSH into the instance.
      external_route_ip: the ip to which traceroute from the VM
      traceroute_args: Additional traceroute args to be passed on.
      dry_run: Whether to only print commands instead of running them.
      resource_registry: gcloud class used for obtaining data from the
        resources.
    Raises:
      ssh.CommandError: there was an error running a SSH command
    """
    instance_string = internal_helpers.GetInstanceNetworkTitleString(instance)
    log.out.Print('<<< Reverse tracerouting from %s' % instance_string)
    # Necessary because the order of commands in the output
    # would be wrong otherwise (the ssh command will output by its own)
    log.out.flush()

    if dry_run:
      external_route_ip = '<SELF-IP>'
    cmd = ['traceroute', external_route_ip]
    if traceroute_args:
      cmd += traceroute_args
    external_helper.RunSSHCommandToInstance(
        command_list=cmd,
        instance=instance,
        user=user,
        args=self._args,
        ssh_helper=self._ssh_helper,
        dry_run=dry_run)
    # This identifier is a simple delimiter of each traceroute run
    if not dry_run:
      log.out.Print('<<<')

  def CheckTraceroute(self, instance, user, dry_run, resource_registry):
    """Checks whether the instance has traceroute in PATH.

    Args:
      instance: Compute Engine VM.
      user: The user to use to SSH into the instance.
      dry_run: Whether to only print commands instead of running them.
      resource_registry: gcloud class used for obtaining data from the
        resources.
    Returns:
      True if the instance has traceroute in PATH,
      False otherwise
    Raises:
      ssh.CommandError: there was an error running a SSH command
    """
    instance_string = internal_helpers.GetInstanceNetworkTitleString(instance)
    log.out.write('Checking traceroute for %s: ' % instance_string)
    if dry_run:
      log.out.Print('[DRY-RUN] No command executed.')
    log.out.flush()

    cmd = ['which', 'traceroute']
    try:
      # This command is silent
      with files.FileWriter(os.devnull) as dev_null:
        return_code = external_helper.RunSSHCommandToInstance(
            command_list=cmd,
            instance=instance,
            user=user,
            args=self._args,
            ssh_helper=self._ssh_helper,
            explicit_output_file=dev_null,
            dry_run=dry_run)
    except Exception as e:
      log.out.write(six.text_type(e))
      log.out.write('\n')  # Close the open print stmt
      log.out.flush()
      raise ssh.CommandError(' '.join(cmd), six.text_type(e))

    if return_code == 0:
      log.out.Print('Traceroute found in PATH')
    else:
      log.out.Print('Traceroute not found in PATH')
    log.out.flush()
    return return_code == 0

  def ObtainSelfIp(self, instance, user, dry_run, resource_registry):
    """Returns the localhost ip as seen from the VM.

    Args:
      instance: Compute Engine VM.
      user: The user to use to SSH into the instance.
      dry_run: Whether to only print commands instead of running them.
      resource_registry: gcloud class used for obtaining data from the
        resources.
    Returns:
      A string containing the local ip,
      None if the obtaining was unsuccessful
    Raises:
      ssh.CommandError: there was an error running a SSH command
    """
    instance_string = internal_helpers.GetInstanceNetworkTitleString(instance)
    log.out.write('Obtaining self ip from %s: ' % instance_string)
    # Sometimes this call will appear after the actual result
    log.out.flush()
    if dry_run:
      log.out.Print('<SELF-IP>')

    temp = io.BytesIO()
    cmd = ['echo', '$SSH_CLIENT']
    try:
      external_helper.RunSSHCommandToInstance(
          command_list=cmd,
          instance=instance,
          user=user,
          args=self._args,
          ssh_helper=self._ssh_helper,
          explicit_output_file=temp,
          dry_run=dry_run)
    except Exception as e:  # pylint: disable=broad-exception
      log.out.write('\n')  # Close the open print stmt
      log.out.flush()
      raise ssh.CommandError(' '.join(cmd), six.text_type(e))

    who_am_i_str = temp.getvalue().decode('utf-8')
    result = re.search(r'(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})', who_am_i_str)
    if result:
      res = result.group(1)
      log.out.Print(res)
      log.out.flush()
      return res

    return None

  @property
  def resource_type(self):
    return 'instances'


class _RoutesArgs(object):
  """Helper to setting and getting values for the args."""

  @classmethod
  def Args(cls, parser):
    """Creates the flags stmts for the command."""
    # Gives us the basic SSH flags
    ssh_utils.BaseSSHCLIHelper.Args(parser)

    base_classes.ZonalLister.Args(parser)

    # SSH flag
    parser.add_argument(
        '--container',
        help="""\
            The name or ID of a container inside of the virtual machine instance
            to connect to. This only applies to virtual machines that are using
            a Container-Optimized OS virtual machine image.
            For more information, see
            [](https://cloud.google.com/compute/docs/containers)
            """)

    parser.add_argument(
        '--external-route-ip',
        default=None,
        help=
        ('For reverse traceroute, this will be the ip given to the VM instance '
         'to traceroute to. This will override all obtained ips.'))

    parser.add_argument(
        '--reverse-traceroute',
        action='store_true',
        help='If enabled, will also run traceroute from the VM to the host')

    # SSH flag
    parser.add_argument(
        '--ssh-flag',
        action='append',
        help="""\
        Additional flags to be passed to *ssh(1)*. It is recommended that flags
        be passed using an assignment operator and quotes. This flag will
        replace occurences of ``%USER%'' and ``%INSTANCE%'' with their
        dereferenced values. Example:

          $ {command} example-instance --zone us-central1-a \
          --ssh-flag="-vvv" --ssh-flag="-L 80:%INSTANCE%:80"

        is equivalent to passing the flags ``--vvv'' and ``-L
        80:162.222.181.197:80'' to *ssh(1)* if the external IP address of
        'example-instance' is 162.222.181.197.
        """)

    parser.add_argument(
        '--user',
        help="""\
        User for login to the selected VMs.
        If not specified, the default user will be used.
        """)

    parser.add_argument(
        'traceroute_args',
        nargs=argparse.REMAINDER,
        help="""\
            Flags and positionals passed to the underlying traceroute call.
            """,
        example="""\
            $ {command} example-instance -- -w 0.5 -q 5 42
        """)

  @classmethod
  def GetFilters(cls, args):
    filters = []
    if args.regexp:
      filters.append('name eq %s' % args.regexp)

    if not filters:
      return None
    filters = ' AND '.join(filters)
    return filters


class _RoutesQueries(object):
  """Helper for getting instance queries using the gcloud SDK."""

  @classmethod
  def ObtainInstances(cls, names, **kwargs):
    """Returns a list of instances according to the flags."""
    errors = []
    result = lister.GetZonalResources(
        service=kwargs['service'],
        project=kwargs['project'],
        requested_zones=kwargs['zones'],
        filter_expr=kwargs['filters'],
        http=kwargs['http'],
        batch_url=kwargs['batch_url'],
        errors=errors)
    instances = list(result)

    # We filter them according to the names
    filtered_instances = []
    if not names:
      filtered_instances = instances
    else:
      for name in names:
        # First compare by name
        name_match = None
        in_name = None
        in_self_link = None

        for instance in instances:
          if name == instance.name:
            # Exact name match has a priority
            # over loose match on instance name or selfLink
            name_match = instance
            break
          elif name in instance.name:
            in_name = instance
          elif name in instance.selfLink:
            in_self_link = instance

        if name_match:
          filtered_instances.append(name_match)
        elif in_name:
          filtered_instances.append(in_name)
        elif in_self_link:
          filtered_instances.append(in_self_link)

    return filtered_instances



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for attaching a persistent disk to a TPU VM.

Allows TPU VM users to attach persistent disks to TPUs
in a form that is decoupled from the Create and Delete
lifecycle of the actual TPU VM.
"""


import re
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.compute.tpus.tpu_vm import resource_args
from googlecloudsdk.command_lib.compute.tpus.tpu_vm import util as tpu_utils
from googlecloudsdk.command_lib.util.concepts import concept_parsers
from googlecloudsdk.core import properties

MODE_OPTIONS = {
    'read-write':
        ('Read-write. It is an error to attach a disk in read-write mode to '
         'more than one TPU VM.'),
    'read-only':
        'Read-only.',
}

DETAILED_HELP = {
    'DESCRIPTION':
        """
        *{command}* is used to attach a disk to a TPU VM. For example,

          $ gcloud compute tpus tpu-vm attach-disk example-tpu --disk=example-disk --mode=read-only --zone=us-central1-a

        attaches the disk named 'example-disk' to a TPU VM named
        'example-tpu' in read-only mode in zone ``us-central1-a''.

        """,
    'EXAMPLES':
        """
        To attach a disk named 'my-disk' for read-only access to a TPU named
        'my-tpu', run:

          $ {command} my-tpu --disk=my-disk --mode=read-only

        To attach a disk named 'my-read-write-disk' for read-write access to
        a TPU named 'my-tpu', run:

          $ {command} my-tpu --disk=my-read-write-disk

        To attach a regional disk named for read-write access to a TPU named
        'my-tpu', run:

          $ {command} my-tpu --disk=projects/tpu-test-env-one-vm/region/us-central1/disks/example-disk
        """,
}


def AddTPUResourceArg(parser, verb):
  """Adds a TPU Name resource argument.

  NOTE: May be used only if it's the only resource arg in the command.

  Args:
    parser: the argparse parser for the command.
    verb: str, the verb to describe the resource, such as 'to attach'.
  """

  concept_parsers.ConceptParser.ForResource(
      'tpu',
      resource_args.GetTPUResourceSpec('TPU'),
      'The TPU function name {}.'.format(verb),
      required=True).AddToParser(parser)


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class AttachDisk(base.Command):
  """Attach disk to TPU VM."""

  @staticmethod
  def Args(parser):
    """Set up arguments for this command.

    Args:
      parser: An argparse.ArgumentParser.
    """

    parser.add_argument(
        '--disk',
        help='The name of the disk to attach to the TPU VM.',
        required=True)

    parser.add_argument(
        '--mode',
        choices=MODE_OPTIONS,
        default='read-write',
        help='Specifies the mode of the disk.')

    AddTPUResourceArg(parser, 'to attach disk')

  def Run(self, args):
    # If zone is not set, retrieve the one from the config.
    if args.zone is None:
      args.zone = properties.VALUES.compute.zone.Get(required=True)

    # Retrieve the TPU node.
    tpu_name_ref = args.CONCEPTS.tpu.Parse()
    tpu = tpu_utils.TPUNode(self.ReleaseTrack())
    node = tpu.Get(tpu_name_ref.Name(), args.zone)

    if not tpu_utils.IsTPUVMNode(node):
      raise exceptions.BadArgumentException(
          'TPU',
          'this command is only available for Cloud TPU VM nodes. To access '
          'this node, please see '
          'https://cloud.google.com/tpu/docs/creating-deleting-tpus.')

    if args.mode == 'read-write':
      args.mode = tpu.messages.AttachedDisk.ModeValueValuesEnum.READ_WRITE
    elif args.mode == 'read-only':
      args.mode = tpu.messages.AttachedDisk.ModeValueValuesEnum.READ_ONLY
    else:
      raise exceptions.BadArgumentException(
          'mode', 'can only attach disks in read-write or read-only mode.'
      )

    source_path_match = re.match(r'projects/.+/(zones|regions)/.+/disks/.+',
                                 args.disk)
    if source_path_match:
      source_path = args.disk
    else:
      project = properties.VALUES.core.project.Get(required=True)
      source_path = (
          'projects/' + project + '/zones/' + args.zone + '/disks/' + args.disk
      )

    if not node.dataDisks:
      disk_to_attach = tpu.messages.AttachedDisk(
          mode=args.mode,
          sourceDisk=source_path
      )
      node.dataDisks.append(disk_to_attach)
    else:
      source_disk_list = []
      for disk in node.dataDisks:
        source_disk_list.append(disk.sourceDisk)
      if source_path not in source_disk_list:
        disk_to_attach = tpu.messages.AttachedDisk(
            mode=args.mode,
            sourceDisk=source_path
        )
        node.dataDisks.append(disk_to_attach)
      else:
        raise exceptions.BadArgumentException(
            'TPU', 'disk is already attached to the TPU VM.')

    return tpu.UpdateNode(
        name=tpu_name_ref.Name(),
        zone=args.zone,
        node=node,
        update_mask='data_disks',
        poller_message='Attaching disk to TPU VM',
    )

AttachDisk.detailed_help = DETAILED_HELP



# Copyright 2016 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Conversions to translate between legacy YAML and OnePlatform protos."""

from __future__ import absolute_import
from __future__ import annotations
import re

# pylint:disable=g-import-not-at-top
try:
  from googlecloudsdk.appengine.api import dispatchinfo
except ImportError:
  from google.appengine.api import dispatchinfo
try:
  from googlecloudsdk.appengine.api import appinfo
except ImportError:
  from google.appengine.api import appinfo
# pylint:enable=g-import-not-at-top

_SECONDS_PER_MINUTE = 60
_MILLISECONDS_PER_SECOND = 1000
_NANOSECONDS_PER_SECOND = 1000000000

_COMMON_HANDLER_FIELDS = (
    'urlRegex',
    'login',
    'authFailAction',
    'securityLevel',
    'redirectHttpResponseCode',
)

_SCRIPT_FIELDS = (
    'scriptPath',
)

_HANDLER_FIELDS = {
    'staticFiles': (
        'path',
        'uploadPathRegex',
        'httpHeaders',
        'expiration',
        'applicationReadable',
        'mimeType',
        'requireMatchingFile',
    ),
    'script': _SCRIPT_FIELDS,
    'apiEndpoint': _SCRIPT_FIELDS,
}

_REQUEST_UTILIZATION_SCALING_FIELDS = (
    'targetRequestCountPerSec',
    'targetConcurrentRequests',
    'targetRequestCountPerSecond',
)

_DISK_UTILIZATION_SCALING_FIELDS = (
    'targetWriteBytesPerSec',
    'targetWriteOpsPerSec',
    'targetReadBytesPerSec',
    'targetReadOpsPerSec',
    'targetWriteBytesPerSecond',
    'targetWriteOpsPerSecond',
    'targetReadBytesPerSecond',
    'targetReadOpsPerSecond',
)

_NETWORK_UTILIZATION_SCALING_FIELDS = (
    'targetSentBytesPerSec',
    'targetSentPacketsPerSec',
    'targetReceivedBytesPerSec',
    'targetReceivedPacketsPerSec',
    'targetSentBytesPerSecond',
    'targetSentPacketsPerSecond',
    'targetReceivedBytesPerSecond',
    'targetReceivedPacketsPerSecond',
)

_ENDPOINTS_ROLLOUT_STRATEGY_VALUES = (
    'fixed',
    'managed',
)
(_ENDPOINTS_UNSPECIFIED_ROLLOUT_STRATEGY_ENUM_VALUE
) = 'UNSPECIFIED_ROLLOUT_STRATEGY'

_STANDARD_SCHEDULER_SETTINGS = (
    'maxInstances',
    'minInstances',
    'targetCpuUtilization',
    'targetThroughputUtilization',
)

_SUBNETWORK_KEY_FIELDS = ('hostProjectId', 'subnet')


# Maps VPC egress setting as specified in app.yaml to their proto enum values.
_VPC_EGRESS_SETTING_MAP = {
    'all-traffic': 'ALL_TRAFFIC',
    'private-ranges-only': 'PRIVATE_IP_RANGES',
}


# Maps bundled service type as specified in app.yaml to their proto enum values.
_BUNDLED_SERVICE_TYPE_ENUM = {
    'app_identity_service': 'BUNDLED_SERVICE_TYPE_APP_IDENTITY_SERVICE',
    'blobstore': 'BUNDLED_SERVICE_TYPE_BLOBSTORE',
    'capability_service': 'BUNDLED_SERVICE_TYPE_CAPABILITY_SERVICE',
    'datastore_v3': 'BUNDLED_SERVICE_TYPE_DATASTORE_V3',
    'deferred': 'BUNDLED_SERVICE_TYPE_DEFERRED',
    'images': 'BUNDLED_SERVICE_TYPE_IMAGES',
    'mail': 'BUNDLED_SERVICE_TYPE_MAIL',
    'memcache': 'BUNDLED_SERVICE_TYPE_MEMCACHE',
    'modules': 'BUNDLED_SERVICE_TYPE_MODULES',
    'namespaces': 'BUNDLED_SERVICE_TYPE_NAMESPACES',
    'ndb': 'BUNDLED_SERVICE_TYPE_NDB',
    'search': 'BUNDLED_SERVICE_TYPE_SEARCH',
    'taskqueue': 'BUNDLED_SERVICE_TYPE_TASKQUEUES',
    'urlfetch': 'BUNDLED_SERVICE_TYPE_URLFETCH',
    'user': 'BUNDLED_SERVICE_TYPE_USERS',
}


def ToBundledServiceTypeEnum(value):
  """Converts a string to a bundled service type.

  Args:
    value: The bundled service name (string).

  Returns:
    The corresponding enum value (string).

  Raises:
    ValueError: If the provided value is not a valid bundled service name.
  """
  if str(value) not in _BUNDLED_SERVICE_TYPE_ENUM:
    raise ValueError(
        f'Value "{value}" is not a valid bundled service name. '
        f'Expected one of: {_BUNDLED_SERVICE_TYPE_ENUM.keys()}'
    )
  return _BUNDLED_SERVICE_TYPE_ENUM[str(value)]


def ToVpcEgressSettingEnum(value):
  """Converts a string to a VPC egress setting."""
  if str(value) not in _VPC_EGRESS_SETTING_MAP:
    raise ValueError(
        'egress_setting must be one of: [%s]'
        % ','.join(_VPC_EGRESS_SETTING_MAP.keys())
    )
  return _VPC_EGRESS_SETTING_MAP[str(value)]


def EnumConverter(prefix):
  """Create conversion function which translates from string to enum value.

  Args:
    prefix: Prefix for enum value. Expected to be an upper-cased value.

  Returns:
    A conversion function which translates from string to enum value.

  Raises:
    ValueError: If an invalid prefix (empty, non-upper-cased, etc.) prefix was
    provided.
  """
  if not prefix:
    raise ValueError('A prefix must be provided')
  if prefix != prefix.upper():
    raise ValueError('Upper-cased prefix must be provided')
  if prefix.endswith('_'):
    raise ValueError(
        'Prefix should not contain a trailing underscore: "%s"' % prefix)

  return lambda value: '_'.join([prefix, str(value).upper()])


def Not(value):
  """Convert the given boolean value to the opposite value."""
  if not isinstance(value, bool):
    raise ValueError('Expected a boolean value. Got "%s"' % value)
  return not value


def ToJsonString(value):
  """Coerces a primitive value into a JSON-compatible string.

  Special handling for boolean values, since the Python version (True/False) is
  incompatible with the JSON version (true/false).

  Args:
    value: value to convert.

  Returns:
    Value as a string.

  Raises:
    ValueError: when a non-primitive value is provided.
  """
  if isinstance(value, (list, dict)):
    raise ValueError('Expected a primitive value. Got "%s"' % value)
  if isinstance(value, bool):
    return str(value).lower()
  return str(value)


def ToUpperCaseJsonString(value):
  """Coerces a primitive value into a upper-case JSON-compatible string.

  Special handling for  values whose JSON version is in upper-case.

  Args:
    value: value to convert.

  Returns:
    Value as a string.

  Raises:
    ValueError: when a non-primitive value is provided.
  """
  return str(value).upper()


def StringToInt(handle_automatic=False):
  """Create conversion function which converts from a string to an integer.

  Args:
    handle_automatic: Boolean indicating whether a value of "automatic" should
      be converted to 0.

  Returns:
    A conversion function which converts a string to an integer.
  """
  def Convert(value):
    if value == 'automatic' and handle_automatic:
      return 0
    return int(value)

  return Convert


def SecondsToDuration(value):
  """Convert seconds expressed as integer to a Duration value."""
  return '%ss' % int(value)


def LatencyToDuration(value):
  """Convert valid pending latency argument to a Duration value of seconds.

  Args:
    value: A string in the form X.Xs or XXms.

  Returns:
    Duration value of the given argument.

  Raises:
    ValueError: if the given value isn't parseable.
  """
  if not re.compile(appinfo._PENDING_LATENCY_REGEX).match(value):  # pylint: disable=protected-access
    raise ValueError('Unrecognized latency: %s' % value)
  if value == 'automatic':
    return None
  if value.endswith('ms'):
    return '%ss' % (float(value[:-2]) / _MILLISECONDS_PER_SECOND)
  else:
    return value


def IdleTimeoutToDuration(value):
  """Convert valid idle timeout argument to a Duration value of seconds.

  Args:
    value: A string in the form Xm or Xs

  Returns:
    Duration value of the given argument.

  Raises:
    ValueError: if the given value isn't parseable.
  """
  if not re.compile(appinfo._IDLE_TIMEOUT_REGEX).match(value):  # pylint: disable=protected-access
    raise ValueError('Unrecognized idle timeout: %s' % value)
  if value.endswith('m'):
    return '%ss' % (int(value[:-1]) * _SECONDS_PER_MINUTE)
  else:
    return value


def ExpirationToDuration(value):
  """Convert valid expiration argument to a Duration value of seconds.

  Args:
    value: String that matches _DELTA_REGEX.

  Returns:
    Time delta expressed as a Duration.

  Raises:
    ValueError: if the given value isn't parseable.
  """
  if not re.compile(appinfo._EXPIRATION_REGEX).match(value):  # pylint: disable=protected-access
    raise ValueError('Unrecognized expiration: %s' % value)
  delta = appinfo.ParseExpiration(value)
  return '%ss' % delta


def ConvertAutomaticScaling(automatic_scaling):
  """Moves several VM-specific automatic scaling parameters to submessages.

  For example:
  Input {
    "targetSentPacketsPerSec": 10,
    "targetReadOpsPerSec": 2,
    "targetRequestCountPerSec": 3
  }
  Output {
    "networkUtilization": {
      "targetSentPacketsPerSec": 10
    },
    "diskUtilization": {
      "targetReadOpsPerSec": 2
    },
    "requestUtilization": {
      "targetRequestCountPerSec": 3
    }
  }

  Args:
    automatic_scaling: Result of converting automatic_scaling according to
      schema.
  Returns:
    AutomaticScaling which has moved network/disk utilization related fields to
    submessage.
  """
  def MoveFieldsTo(field_names, target_field_name):
    target = {}
    for field_name in field_names:
      if field_name in automatic_scaling:
        target[field_name] = automatic_scaling[field_name]
        del automatic_scaling[field_name]
    if target:
      automatic_scaling[target_field_name] = target
  MoveFieldsTo(_REQUEST_UTILIZATION_SCALING_FIELDS, 'requestUtilization')
  MoveFieldsTo(_DISK_UTILIZATION_SCALING_FIELDS, 'diskUtilization')
  MoveFieldsTo(_NETWORK_UTILIZATION_SCALING_FIELDS, 'networkUtilization')
  MoveFieldsTo(_STANDARD_SCHEDULER_SETTINGS, 'standardSchedulerSettings')
  return automatic_scaling


def ConvertUrlHandler(handler):
  """Rejiggers the structure of the url handler based on its type.

  An extra level of message nesting occurs here, based on the handler type.
  Fields common to all handler types occur at the top-level, while
  handler-specific fields will go into a submessage based on handler type.

  For example, a static files handler is transformed as follows:
  Input {
    "urlRegex": "foo/bar.html",
    "path": "static_files/foo/bar.html"
  }
  Output {
    "urlRegex": "foo/bar.html",
    "staticFiles": {
      "path": "static_files/foo/bar.html"
    }
  }

  Args:
    handler: Result of converting handler according to schema.

  Returns:
    Handler which has moved fields specific to the handler's type to a
    submessage.
  """

  def AppendRegexToPath(path, regex):
    """Equivalent to os.path.join(), except uses forward slashes always."""
    return path.rstrip('/') + '/' + regex

  handler_type = _GetHandlerType(handler)

  # static_dir is syntactic sugar for static_files, so we "demote" any
  # static_dir directives we see to a static_files directive before
  # continuing.
  if handler_type == 'staticDirectory':
    # Groups are disallowed in URLs for static directory handlers.
    # We check for them using the Python re module. App Engine uses Posix
    # extended regular expressions, but it's overkill to start packaging a
    # library that officially supports Posix extended regular expressions for
    # this simple validation. We just let compile errors slide; Python regular
    # expressions are mostly a superset.
    try:
      compiled = re.compile(handler['urlRegex'])
    except re.error:
      pass  # We'll let the API handle this.
    else:
      if compiled.groups:  # `groups` is the number of groups in the RE
        raise ValueError(
            'Groups are not allowed in URLs for static directory handlers: ' +
            handler['urlRegex'])
    tmp = {
        'path': AppendRegexToPath(handler['staticDir'], r'\1'),
        'uploadPathRegex': AppendRegexToPath(handler['staticDir'], '.*'),
        'urlRegex': AppendRegexToPath(handler['urlRegex'], '(.*)'),
    }
    del handler['staticDir']
    handler.update(tmp)
    handler_type = 'staticFiles'

  new_handler = {}
  new_handler[handler_type] = {}

  for field in _HANDLER_FIELDS[handler_type]:
    if field in handler:
      new_handler[handler_type][field] = handler[field]

  # Copy the common fields
  for common_field in _COMMON_HANDLER_FIELDS:
    if common_field in handler:
      new_handler[common_field] = handler[common_field]

  return new_handler


def ConvertDispatchHandler(handler):
  """Create conversion function which handles dispatch rules.

  Extract domain and path from dispatch url,
  set service value from service or module info.

  Args:
    handler: Result of converting handler according to schema.

  Returns:
    Handler which has 'domain', 'path' and 'service' fields.
  """
  dispatch_url = dispatchinfo.ParsedURL(handler['url'])
  dispatch_service = handler['service']

  dispatch_domain = dispatch_url.host
  if not dispatch_url.host_exact:
    dispatch_domain = '*' + dispatch_domain

  dispatch_path = dispatch_url.path
  if not dispatch_url.path_exact:
    dispatch_path = dispatch_path.rstrip('/') + '/*'

  new_handler = {}
  new_handler['domain'] = dispatch_domain
  new_handler['path'] = dispatch_path
  new_handler['service'] = dispatch_service

  return new_handler


def _GetHandlerType(handler):
  """Get handler type of mapping.

  Args:
    handler: Original handler.

  Returns:
    Handler type determined by which handler id attribute is set.

  Raises:
    ValueError: when none of the handler id attributes are set.
  """
  if 'apiEndpoint' in handler:
    return 'apiEndpoint'
  elif 'staticDir' in handler:
    return 'staticDirectory'
  elif 'path' in handler:
    return 'staticFiles'
  elif 'scriptPath' in handler:
    return 'script'

  raise ValueError('Unrecognized handler type: %s' % handler)


def ConvertEndpointsRolloutStrategyToEnum(value):
  """Converts the rollout strategy to an enum.

  In the YAML file, the user does not use the enum values directly. Therefore we
  must convert these to their corresponding enum values in version.proto.

  Args:
    value: A string that is a valid rollout strategy ("fixed" or "managed")

  Returns:
    Value converted to the proper enum value. Defaults to
    "UNSPECIFIED_ROLLOUT_STRATEGY"

  Raises:
    ValueError: When the value is set and is not one of "fixed" or "managed".
  """
  if value is None:
    return _ENDPOINTS_UNSPECIFIED_ROLLOUT_STRATEGY_ENUM_VALUE
  if value in _ENDPOINTS_ROLLOUT_STRATEGY_VALUES:
    return value.upper()

  raise ValueError('Unrecognized rollout strategy: %s' % value)


def ConvertEntrypoint(entrypoint):
  """Converts the raw entrypoint to a nested shell value.

  In the YAML file, the user specifies an entrypoint value. However, the version
  resource expects it to be nested under a 'shell' key. In addition, Zeus
  always prepends 'exec' to the value provided, so we remove it here as it is
  sometimes added client-side by the validation library.

  Args:
    entrypoint: string, entrypoint value.

  Returns:
    Dict containing entrypoint.
  """
  if entrypoint is None:
    entrypoint = ''
  if entrypoint.startswith('exec '):
    entrypoint = entrypoint[len('exec '):]
  return {'shell': entrypoint}


def CommaSeparatedStringToList(input_str: str | None) -> list[str]:
  """Converts a comma-separated string to a list of strings.

  Args:
    input_str: A string containing one or more tags, separated by commas, or
      None. Handles empty strings and None by returning an empty list.

  Returns:
    A list of strings. The returned strings are stripped and empty strings are
    filtered from the list.
  """
  if not input_str:
    return []
  return [v.strip() for v in input_str.split(',') if v.strip()]



# -*- coding: utf-8 -*- #
# Copyright 2023 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Argument processors specifically for storage pools."""



GB = 2**30


def convert_to_gb(size_bytes: int) -> int:
  """Convert size_bytes to GB (2^30).

  Args:
    size_bytes: Size in bytes.

  Returns:
    Size in gibibytes.
  """
  # Converting to decimal to not have any precision issues. Doing so is
  # computationally not cheap, though it is fine for invoking it once
  # for an argument.
  return size_bytes // GB



# -*- coding: utf-8 -*- #
# Copyright 2016 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to list all billing accounts associated with the active user."""


from googlecloudsdk.api_lib.billing import billing_client
from googlecloudsdk.calliope import base
from googlecloudsdk.core import resources


class List(base.ListCommand):
  """List all active billing accounts.

  `{command}` lists all billing accounts and subaccounts owned by the currently
  authenticated user. Subaccounts have a non-empty MASTER_ACCOUNT_ID value.

  ## EXAMPLES

  To list only open billing accounts, run:

      $ {command} --filter=open=true

  ## API REFERENCE

  This command uses the *cloudbilling/v1* API. The full documentation for this
  API can be found at: https://cloud.google.com/billing/v1/getting-started
  """

  @staticmethod
  def ToSelfLink(account):
    return resources.REGISTRY.Parse(
        account.name, collection='cloudbilling.billingAccounts').SelfLink()

  @staticmethod
  def Args(parser):
    parser.display_info.AddFormat("""
          table(
            name.basename():label=ACCOUNT_ID,
            displayName:label=NAME,
            open,
            masterBillingAccount.basename():label=MASTER_ACCOUNT_ID
          )
    """)
    parser.display_info.AddUriFunc(List.ToSelfLink)

  def Run(self, args):
    """Run the list command."""
    client = billing_client.AccountsClient()
    return client.List(limit=args.limit)



# -*- coding: utf-8 -*- #
# Copyright 2024 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for the Resource Settings service."""


from googlecloudsdk.api_lib.util import apis

RESOURCE_SETTINGS_API_NAME = 'cloudresourcemanager'
RESOURCE_SETTINGS_API_VERSION = 'v3'


def ResourceSettingsClient():
  """Returns a client instance of the Resource Settings service."""
  return apis.GetClientInstance(RESOURCE_SETTINGS_API_NAME,
                                RESOURCE_SETTINGS_API_VERSION)


def ResourceSettingsMessages():
  """Returns the messages module for the Resource Settings service."""
  return apis.GetMessagesModule(RESOURCE_SETTINGS_API_NAME,
                                RESOURCE_SETTINGS_API_VERSION)


# Folders Settings
def FoldersSettingsService():
  """Returns the service class for the Folders Settings resource."""
  client = ResourceSettingsClient()
  return client.folders_settings


# Folders Effective Settings
def FoldersEffectiveSettingsService():
  """Returns the service class for the Folders Effective Settings resource."""
  client = ResourceSettingsClient()
  return client.folders_effectiveSettings


# Organization Settings
def OrganizationsSettingsService():
  """Returns the service class for the Organization Settings resource."""
  client = ResourceSettingsClient()
  return client.organizations_settings


# Organization Effective Settings
def OrganizationsEffectiveSettingsService():
  """Returns the service class for the Organization Effective Settings resource."""
  client = ResourceSettingsClient()
  return client.organizations_effectiveSettings


# Project Settings
def ProjectsSettingsService():
  """Returns the service class for the Project Settings resource."""
  client = ResourceSettingsClient()
  return client.projects_settings


# Project Effective Settings
def ProjectsEffectiveSettingsService():
  """Returns the service class for the Project Effective Settings resource."""
  client = ResourceSettingsClient()
  return client.projects_effectiveSettings



"""Generated client library for biglake version v1beta."""
# NOTE: This file is autogenerated and should not be edited by hand.

from apitools.base.py import base_api
from googlecloudsdk.generated_clients.apis.biglake.v1beta import biglake_v1beta_messages as messages


class BiglakeV1beta(base_api.BaseApiClient):
  """Generated client library for service biglake version v1beta."""

  MESSAGES_MODULE = messages
  BASE_URL = 'https://biglake.googleapis.com/'
  MTLS_BASE_URL = 'https://biglake.mtls.googleapis.com/'

  _PACKAGE = 'biglake'
  _SCOPES = ['https://www.googleapis.com/auth/bigquery', 'https://www.googleapis.com/auth/cloud-platform']
  _VERSION = 'v1beta'
  _CLIENT_ID = 'CLIENT_ID'
  _CLIENT_SECRET = 'CLIENT_SECRET'
  _USER_AGENT = 'google-cloud-sdk'
  _CLIENT_CLASS_NAME = 'BiglakeV1beta'
  _URL_VERSION = 'v1beta'
  _API_KEY = None

  def __init__(self, url='', credentials=None,
               get_credentials=True, http=None, model=None,
               log_request=False, log_response=False,
               credentials_args=None, default_global_params=None,
               additional_http_headers=None, response_encoding=None):
    """Create a new biglake handle."""
    url = url or self.BASE_URL
    super(BiglakeV1beta, self).__init__(
        url, credentials=credentials,
        get_credentials=get_credentials, http=http, model=model,
        log_request=log_request, log_response=log_response,
        credentials_args=credentials_args,
        default_global_params=default_global_params,
        additional_http_headers=additional_http_headers,
        response_encoding=response_encoding)
    self.hive_v1beta_projects_catalogs_databases_tables_partitions = self.HiveV1betaProjectsCatalogsDatabasesTablesPartitionsService(self)
    self.hive_v1beta_projects_catalogs_databases_tables = self.HiveV1betaProjectsCatalogsDatabasesTablesService(self)
    self.hive_v1beta_projects_catalogs_databases = self.HiveV1betaProjectsCatalogsDatabasesService(self)
    self.hive_v1beta_projects_catalogs = self.HiveV1betaProjectsCatalogsService(self)
    self.hive_v1beta_projects = self.HiveV1betaProjectsService(self)
    self.hive_v1beta = self.HiveV1betaService(self)
    self.hive = self.HiveService(self)

  class HiveV1betaProjectsCatalogsDatabasesTablesPartitionsService(base_api.BaseApiService):
    """Service class for the hive_v1beta_projects_catalogs_databases_tables_partitions resource."""

    _NAME = 'hive_v1beta_projects_catalogs_databases_tables_partitions'

    def __init__(self, client):
      super(BiglakeV1beta.HiveV1betaProjectsCatalogsDatabasesTablesPartitionsService, self).__init__(client)
      self._upload_configs = {
          }

    def BatchCreate(self, request, global_params=None):
      r"""Adds partitions to a table.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesTablesPartitionsBatchCreateRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (BatchCreatePartitionsResponse) The response message.
      """
      config = self.GetMethodConfig('BatchCreate')
      return self._RunMethod(
          config, request, global_params=global_params)

    BatchCreate.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}/tables/{tablesId}/partitions:batchCreate',
        http_method='POST',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.tables.partitions.batchCreate',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=[],
        relative_path='hive/v1beta/{+parent}/partitions:batchCreate',
        request_field='batchCreatePartitionsRequest',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesTablesPartitionsBatchCreateRequest',
        response_type_name='BatchCreatePartitionsResponse',
        supports_download=False,
    )

    def BatchDelete(self, request, global_params=None):
      r"""Deletes partitions from a table.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesTablesPartitionsBatchDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('BatchDelete')
      return self._RunMethod(
          config, request, global_params=global_params)

    BatchDelete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}/tables/{tablesId}/partitions:batchDelete',
        http_method='POST',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.tables.partitions.batchDelete',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=[],
        relative_path='hive/v1beta/{+parent}/partitions:batchDelete',
        request_field='batchDeletePartitionsRequest',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesTablesPartitionsBatchDeleteRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def BatchUpdate(self, request, global_params=None):
      r"""Updates partitions in a table.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesTablesPartitionsBatchUpdateRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (BatchUpdatePartitionsResponse) The response message.
      """
      config = self.GetMethodConfig('BatchUpdate')
      return self._RunMethod(
          config, request, global_params=global_params)

    BatchUpdate.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}/tables/{tablesId}/partitions:batchUpdate',
        http_method='POST',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.tables.partitions.batchUpdate',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=[],
        relative_path='hive/v1beta/{+parent}/partitions:batchUpdate',
        request_field='batchUpdatePartitionsRequest',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesTablesPartitionsBatchUpdateRequest',
        response_type_name='BatchUpdatePartitionsResponse',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Streams list of partitions from a table.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesTablesPartitionsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListPartitionsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}/tables/{tablesId}/partitions:list',
        http_method='GET',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.tables.partitions.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['filter'],
        relative_path='hive/v1beta/{+parent}/partitions:list',
        request_field='',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesTablesPartitionsListRequest',
        response_type_name='ListPartitionsResponse',
        supports_download=False,
    )

  class HiveV1betaProjectsCatalogsDatabasesTablesService(base_api.BaseApiService):
    """Service class for the hive_v1beta_projects_catalogs_databases_tables resource."""

    _NAME = 'hive_v1beta_projects_catalogs_databases_tables'

    def __init__(self, client):
      super(BiglakeV1beta.HiveV1betaProjectsCatalogsDatabasesTablesService, self).__init__(client)
      self._upload_configs = {
          }

    def Create(self, request, global_params=None):
      r"""Creates a new hive table.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesTablesCreateRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (HiveTable) The response message.
      """
      config = self.GetMethodConfig('Create')
      return self._RunMethod(
          config, request, global_params=global_params)

    Create.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}/tables',
        http_method='POST',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.tables.create',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['hiveTableId'],
        relative_path='hive/v1beta/{+parent}/tables',
        request_field='hiveTable',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesTablesCreateRequest',
        response_type_name='HiveTable',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes an existing table specified by the table name.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesTablesDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}/tables/{tablesId}',
        http_method='DELETE',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.tables.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='hive/v1beta/{+name}',
        request_field='',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesTablesDeleteRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets the table specified by the resource name.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesTablesGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (HiveTable) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}/tables/{tablesId}',
        http_method='GET',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.tables.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='hive/v1beta/{+name}',
        request_field='',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesTablesGetRequest',
        response_type_name='HiveTable',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""List all hive tables in a specified project under the hive catalog and database.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesTablesListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListHiveTablesResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}/tables',
        http_method='GET',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.tables.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['pageSize', 'pageToken'],
        relative_path='hive/v1beta/{+parent}/tables',
        request_field='',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesTablesListRequest',
        response_type_name='ListHiveTablesResponse',
        supports_download=False,
    )

    def Patch(self, request, global_params=None):
      r"""Updates an existing table specified by the table name.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesTablesPatchRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (HiveTable) The response message.
      """
      config = self.GetMethodConfig('Patch')
      return self._RunMethod(
          config, request, global_params=global_params)

    Patch.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}/tables/{tablesId}',
        http_method='PATCH',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.tables.patch',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['updateMask'],
        relative_path='hive/v1beta/{+name}',
        request_field='hiveTable',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesTablesPatchRequest',
        response_type_name='HiveTable',
        supports_download=False,
    )

  class HiveV1betaProjectsCatalogsDatabasesService(base_api.BaseApiService):
    """Service class for the hive_v1beta_projects_catalogs_databases resource."""

    _NAME = 'hive_v1beta_projects_catalogs_databases'

    def __init__(self, client):
      super(BiglakeV1beta.HiveV1betaProjectsCatalogsDatabasesService, self).__init__(client)
      self._upload_configs = {
          }

    def Create(self, request, global_params=None):
      r"""Creates a new database.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesCreateRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (HiveDatabase) The response message.
      """
      config = self.GetMethodConfig('Create')
      return self._RunMethod(
          config, request, global_params=global_params)

    Create.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases',
        http_method='POST',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.create',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['hiveDatabaseId'],
        relative_path='hive/v1beta/{+parent}/databases',
        request_field='hiveDatabase',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesCreateRequest',
        response_type_name='HiveDatabase',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes an existing database specified by the database name.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}',
        http_method='DELETE',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='hive/v1beta/{+name}',
        request_field='',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesDeleteRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets the database specified by the resource name.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (HiveDatabase) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}',
        http_method='GET',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='hive/v1beta/{+name}',
        request_field='',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesGetRequest',
        response_type_name='HiveDatabase',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""List all databases in a specified catalog.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListHiveDatabasesResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases',
        http_method='GET',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['pageSize', 'pageToken'],
        relative_path='hive/v1beta/{+parent}/databases',
        request_field='',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesListRequest',
        response_type_name='ListHiveDatabasesResponse',
        supports_download=False,
    )

    def Patch(self, request, global_params=None):
      r"""Updates an existing database specified by the database name.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDatabasesPatchRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (HiveDatabase) The response message.
      """
      config = self.GetMethodConfig('Patch')
      return self._RunMethod(
          config, request, global_params=global_params)

    Patch.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}/databases/{databasesId}',
        http_method='PATCH',
        method_id='biglake.hive.v1beta.projects.catalogs.databases.patch',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['updateMask'],
        relative_path='hive/v1beta/{+name}',
        request_field='hiveDatabase',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDatabasesPatchRequest',
        response_type_name='HiveDatabase',
        supports_download=False,
    )

  class HiveV1betaProjectsCatalogsService(base_api.BaseApiService):
    """Service class for the hive_v1beta_projects_catalogs resource."""

    _NAME = 'hive_v1beta_projects_catalogs'

    def __init__(self, client):
      super(BiglakeV1beta.HiveV1betaProjectsCatalogsService, self).__init__(client)
      self._upload_configs = {
          }

    def Create(self, request, global_params=None):
      r"""Creates a new hive catalog.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsCreateRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (HiveCatalog) The response message.
      """
      config = self.GetMethodConfig('Create')
      return self._RunMethod(
          config, request, global_params=global_params)

    Create.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs',
        http_method='POST',
        method_id='biglake.hive.v1beta.projects.catalogs.create',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['hiveCatalogId', 'primary_location'],
        relative_path='hive/v1beta/{+parent}/catalogs',
        request_field='hiveCatalog',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsCreateRequest',
        response_type_name='HiveCatalog',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes an existing catalog specified by the catalog ID. Delete will fail if the catalog is not empty.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}',
        http_method='DELETE',
        method_id='biglake.hive.v1beta.projects.catalogs.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='hive/v1beta/{+name}',
        request_field='',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsDeleteRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets the catalog specified by the resource name.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (HiveCatalog) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}',
        http_method='GET',
        method_id='biglake.hive.v1beta.projects.catalogs.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='hive/v1beta/{+name}',
        request_field='',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsGetRequest',
        response_type_name='HiveCatalog',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""List all catalogs in a specified project.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ListHiveCatalogsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs',
        http_method='GET',
        method_id='biglake.hive.v1beta.projects.catalogs.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['pageSize', 'pageToken'],
        relative_path='hive/v1beta/{+parent}/catalogs',
        request_field='',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsListRequest',
        response_type_name='ListHiveCatalogsResponse',
        supports_download=False,
    )

    def Patch(self, request, global_params=None):
      r"""Updates an existing catalog.

      Args:
        request: (BiglakeHiveV1betaProjectsCatalogsPatchRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (HiveCatalog) The response message.
      """
      config = self.GetMethodConfig('Patch')
      return self._RunMethod(
          config, request, global_params=global_params)

    Patch.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='hive/v1beta/projects/{projectsId}/catalogs/{catalogsId}',
        http_method='PATCH',
        method_id='biglake.hive.v1beta.projects.catalogs.patch',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['updateMask'],
        relative_path='hive/v1beta/{+name}',
        request_field='hiveCatalog',
        request_type_name='BiglakeHiveV1betaProjectsCatalogsPatchRequest',
        response_type_name='HiveCatalog',
        supports_download=False,
    )

  class HiveV1betaProjectsService(base_api.BaseApiService):
    """Service class for the hive_v1beta_projects resource."""

    _NAME = 'hive_v1beta_projects'

    def __init__(self, client):
      super(BiglakeV1beta.HiveV1betaProjectsService, self).__init__(client)
      self._upload_configs = {
          }

  class HiveV1betaService(base_api.BaseApiService):
    """Service class for the hive_v1beta resource."""

    _NAME = 'hive_v1beta'

    def __init__(self, client):
      super(BiglakeV1beta.HiveV1betaService, self).__init__(client)
      self._upload_configs = {
          }

  class HiveService(base_api.BaseApiService):
    """Service class for the hive resource."""

    _NAME = 'hive'

    def __init__(self, client):
      super(BiglakeV1beta.HiveService, self).__init__(client)
      self._upload_configs = {
          }



# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for migrate from legacy firewall rules to network firewall policies."""


import itertools
import json
import re

from apitools.base.py import list_pager
from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.api_lib.compute.operations import poller
from googlecloudsdk.api_lib.resource_manager import tags as rm_tags
from googlecloudsdk.api_lib.util import waiter
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.compute.network_firewall_policies import convert_terraform
from googlecloudsdk.command_lib.compute.network_firewall_policies import secure_tags_utils
from googlecloudsdk.command_lib.compute.networks import flags as network_flags
from googlecloudsdk.command_lib.resource_manager import endpoint_utils as endpoints
from googlecloudsdk.command_lib.resource_manager import operations
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core.util import files


def _GetFirewallPoliciesAssociatedWithNetwork(network, firewall_policies):
  filtered_policies = []
  for firewall_policy in firewall_policies:
    associated = False
    for association in firewall_policy.associations:
      if association.attachmentTarget == network.selfLink:
        associated = True
    if associated:
      filtered_policies.append(firewall_policy)
  return filtered_policies


def _GetFirewallsAssociatedWithNetwork(network, firewalls):
  filtered_firewalls = []
  for firewall in firewalls:
    if firewall.network == network.selfLink:
      filtered_firewalls.append(firewall)
  return filtered_firewalls


def _GetLegacyTags(selected_firewalls):
  tags = set()
  for firewall in selected_firewalls:
    tags.update(firewall.sourceTags)
    tags.update(firewall.targetTags)
  return tags


def _GetServiceAccounts(selected_firewalls, keep_target_service_accounts):
  service_accounts = set()
  for firewall in selected_firewalls:
    service_accounts.update(firewall.sourceServiceAccounts)
    if not keep_target_service_accounts:
      service_accounts.update(firewall.targetServiceAccounts)
  return service_accounts


def _IsDefaultFirewallPolicyRule(rule):
  # Default egress/ingress IPv4/IPv6 rules
  if 2147483644 <= rule.priority <= 2147483647:
    return True
  # Probably a user defined rule
  return False


def _UnsupportedTagResult(field, tag):
  return (False, "Mapping for {} '{}' was not found.".format(field, tag))


def _IsFirewallSupported(firewall, tag_mapping, keep_target_service_accounts):
  """Checks if the given VPC Firewall can be converted by the Migration Tool."""
  # Source Service Accounts
  for service_account in firewall.sourceServiceAccounts:
    prefixed_service_account = 'sa:' + service_account
    if prefixed_service_account not in tag_mapping:
      return _UnsupportedTagResult(
          'source_service_account', prefixed_service_account
      )
  # Target Service Accounts
  if not keep_target_service_accounts:
    for service_account in firewall.targetServiceAccounts:
      prefixed_service_account = 'sa:' + service_account
      if prefixed_service_account not in tag_mapping:
        return _UnsupportedTagResult(
            'target_service_account', prefixed_service_account
        )
  # Source Tags
  for tag in firewall.sourceTags:
    if tag not in tag_mapping:
      return _UnsupportedTagResult('source_tag', tag)
  # Target Tags
  for tag in firewall.targetTags:
    if tag not in tag_mapping:
      return _UnsupportedTagResult('target_tag', tag)
  return (True, '')


def _IsExcludedFirewall(firewall, patterns):
  for pattern in patterns:
    if re.match(pattern, firewall.name):
      return True
  return False


def _ConvertRuleDirection(messages, direction):
  if direction == messages.Firewall.DirectionValueValuesEnum.INGRESS:
    return messages.FirewallPolicyRule.DirectionValueValuesEnum.INGRESS
  return messages.FirewallPolicyRule.DirectionValueValuesEnum.EGRESS


def _ConvertLayer4Configs(messages, l4_configs):
  layer4_configs = []
  for config in l4_configs:
    layer4_configs.append(
        messages.FirewallPolicyRuleMatcherLayer4Config(
            ipProtocol=config.IPProtocol, ports=config.ports
        )
    )
  return layer4_configs


def _ConvertTags(messages, tag_mapping, tags):
  return [
      messages.FirewallPolicyRuleSecureTag(name=tag_mapping[tag])
      for tag in tags
  ]


def _ConvertServiceAccounts(messages, tag_mapping, service_accounts):
  return [
      messages.FirewallPolicyRuleSecureTag(
          name=tag_mapping['sa:' + service_account]
      )
      for service_account in service_accounts
  ]


def _ConvertRuleInternal(
    messages,
    firewall,
    action,
    l4_configs,
    tag_mapping,
    keep_target_service_accounts,
):
  """Converts VPC Firewall to FirewallPolicy.Rule."""
  target_service_accounts = firewall.targetServiceAccounts
  target_secure_tags = _ConvertTags(messages, tag_mapping, firewall.targetTags)
  if not keep_target_service_accounts:
    target_service_accounts = []
    target_secure_tags = target_secure_tags + _ConvertServiceAccounts(
        messages, tag_mapping, firewall.targetServiceAccounts
    )
  return messages.FirewallPolicyRule(
      disabled=firewall.disabled,
      ruleName=firewall.name,  # Allow and deny cannot be in the same rule
      description=firewall.description,  # Do not change description
      direction=_ConvertRuleDirection(messages, firewall.direction),
      priority=firewall.priority,
      action=action,
      enableLogging=firewall.logConfig.enable,
      match=messages.FirewallPolicyRuleMatcher(
          destIpRanges=firewall.destinationRanges,
          srcIpRanges=firewall.sourceRanges,
          srcSecureTags=(
              _ConvertTags(messages, tag_mapping, firewall.sourceTags)
              + _ConvertServiceAccounts(
                  messages, tag_mapping, firewall.sourceServiceAccounts
              )
          ),
          layer4Configs=_ConvertLayer4Configs(messages, l4_configs),
      ),
      targetServiceAccounts=target_service_accounts,
      targetSecureTags=target_secure_tags,
  )


def _ConvertRule(messages, firewall, tag_mapping, keep_target_service_accounts):
  if firewall.denied:
    return _ConvertRuleInternal(
        messages,
        firewall,
        'deny',
        firewall.denied,
        tag_mapping,
        keep_target_service_accounts,
    )
  return _ConvertRuleInternal(
      messages,
      firewall,
      'allow',
      firewall.allowed,
      tag_mapping,
      keep_target_service_accounts,
  )


def _IsPrefixTrue(statuses):
  false_detected = False
  for status in statuses:
    if status and false_detected:
      return False
    false_detected = false_detected or not status
  return True


def _IsSuffixTrue(statuses):
  statuses_copy = statuses
  statuses_copy.reverse()
  return _IsPrefixTrue(statuses_copy)


def _ReadTagMapping(file_name):
  """Imports legacy to secure tag mapping from a JSON file."""
  try:
    with files.FileReader(file_name) as f:
      data = json.load(f)
  except FileNotFoundError:
    log.status.Print(
        "File '{file}' was not found. Tag mapping was not imported.".format(
            file=file_name
        )
    )
    return None
  except OSError:
    log.status.Print(
        "OS error occurred when opening the file '{file}'. Tag mapping was not"
        ' imported.'.format(file=file_name)
    )
    return None
  except Exception as e:  # pylint: disable=broad-except
    log.status.Print(
        "Unexpected error occurred when reading the JSON file '{file}'. Tag"
        ' mapping was not imported.'.format(file=file_name)
    )
    log.status.Print(repr(e))
    return None

  tag_mapping = {
      k: secure_tags_utils.TranslateSecureTag(v) for k, v in data.items()
  }

  return tag_mapping


def _GetFullCanonicalResourceName(instance):
  start_index = instance.selfLink.find('/projects/')
  resource_name = '//compute.googleapis.com' + instance.selfLink[start_index:]
  return resource_name.replace(
      'instances/%s' % instance.name,
      'instances/%s' % instance.id,
  )


def _GetInstanceLocation(instance):
  return instance.zone[instance.zone.find('/zones/') + len('/zones/') :]


def _GetInstancesInNetwork(project, network_name, compute_client):
  """Gets instances in the network."""

  def _HasInterfaceMatchingNetwork(instance):
    return len([
        network_interface.network
        for network_interface in instance.networkInterfaces
        if network_interface.network.endswith('/%s' % network_name)
    ])

  messages = compute_client.MESSAGES_MODULE
  instance_aggregations = compute_client.instances.AggregatedList(
      messages.ComputeInstancesAggregatedListRequest(
          project=project,
          includeAllScopes=True,
          maxResults=500,
      )
  )

  instances_list = [
      item.value.instances
      for item in instance_aggregations.items.additionalProperties
  ]

  # flatten the list
  instances = list(itertools.chain(*instances_list))
  return list(filter(_HasInterfaceMatchingNetwork, instances))


def _BindTagToInstance(tag_value, instance):
  """Binds tag to the instance."""
  messages = rm_tags.TagMessages()
  resource_name = _GetFullCanonicalResourceName(instance)

  tag_binding = messages.TagBinding(parent=resource_name, tagValue=tag_value)
  binding_req = messages.CloudresourcemanagerTagBindingsCreateRequest(
      tagBinding=tag_binding
  )

  location = _GetInstanceLocation(instance)

  with endpoints.CrmEndpointOverrides(location):
    try:
      op = rm_tags.TagBindingsService().Create(binding_req)
      if not op.done:
        operations.WaitForReturnOperation(
            op,
            'Waiting for TagBinding for parent [{}] and tag value [{}] to be '
            'created with [{}]'.format(resource_name, tag_value, op.name),
        )
    except Exception as e:  # pylint: disable=broad-except
      log.status.Print('Tag binding could not be created: ' + repr(e))


def _BindSecureTagsToInstances(
    network_name, project, tag_mapping_file_name, compute_client
):
  """Binds secure tags to instances with matching network tags."""
  tag_mapping = _ReadTagMapping(tag_mapping_file_name)
  if not tag_mapping:
    return

  vm_instances = _GetInstancesInNetwork(project, network_name, compute_client)

  for vm in vm_instances:
    _BindTagsToInstance(tag_mapping, vm)
    _BindServiceTagsToInstance(tag_mapping, vm)


def _BindTagsToInstance(tag_mapping, vm):
  for tag in vm.tags.items:
    if tag in tag_mapping:
      _BindTagToInstance(tag_mapping[tag], vm)


def _BindServiceTagsToInstance(tag_mapping, vm):
  service_accounts = [sa.email for sa in vm.serviceAccounts]

  for sa in service_accounts:
    prefixed_tag = 'sa:' + sa
    if prefixed_tag in tag_mapping:
      _BindTagToInstance(tag_mapping[prefixed_tag], vm)


def _WriteTagMapping(file_name, tags, service_accounts):
  """Exports legacy to secure tag mapping to a JSON file."""
  # Prefix service account tags with 'sa:'
  prefixed_service_accounts = set(map(lambda x: ('sa:' + x), service_accounts))
  mapping = dict.fromkeys(tags.union(prefixed_service_accounts))

  try:
    with files.FileWriter(path=file_name, create_path=True) as f:
      json.dump(mapping, f)
  except OSError:
    log.status.Print(
        "OS error occurred when opening the file '{file}'. Tag mapping was not"
        ' exported.'.format(file=file_name)
    )
    return
  except Exception as e:  # pylint: disable=broad-except
    log.status.Print(
        "Unexpected error occurred when writing the JSON file '{file}'. Tag"
        ' mapping was not exported.'.format(file=file_name)
    )
    log.status.Print(repr(e))


def _WriteTerraformScript(file_name, tf_script):
  """Exports Terraform script."""

  try:
    with files.FileWriter(path=file_name, create_path=True) as f:
      f.write(tf_script)
  except OSError:
    log.status.Print(
        "OS error occurred when opening the file '{file}'. Terraform script was"
        ' not exported.'.format(file=file_name)
    )
    return
  except Exception as e:  # pylint: disable=broad-except
    log.status.Print(
        "Unexpected error occurred when writing to the file '{file}'. Terraform"
        ' script was not exported.'.format(file=file_name)
    )
    log.status.Print(repr(e))


def _WriteExclusionPatterns(file_name, patterns):
  """Exports regexes used for excluding firewalls."""

  try:
    with files.FileWriter(path=file_name, create_path=True) as f:
      for pattern in patterns:
        f.write(pattern + '\n')
  except OSError:
    log.status.Print(
        "OS error occurred when opening the file '{file}'. Exclusion patterns"
        ' were not exported.'.format(file=file_name)
    )
    return
  except Exception as e:  # pylint: disable=broad-except
    log.status.Print(
        "Unexpected error occurred when writing to the file '{file}'. Exclusion"
        ' patterns were not exported.'.format(file=file_name)
    )
    log.status.Print(repr(e))


def _ReadExclusionPatterns(file_name):
  """Imports exclusion patterns from a file."""
  try:
    with files.FileReader(file_name) as f:
      lines = f.readlines()
      patterns = [line.rstrip('\n') for line in lines]
  except FileNotFoundError:
    log.status.Print(
        "File '{file}' was not found. Exclusion patterns were not imported."
        .format(file=file_name)
    )
    return [], True
  except OSError:
    log.status.Print(
        "OS error occurred when opening the file '{file}'. Exclusion patterns"
        ' were not imported.'.format(file=file_name)
    )
    return [], True
  except Exception as e:  # pylint: disable=broad-except
    log.status.Print(
        "Unexpected error occurred when reading the file '{file}'. Exclusion"
        ' patterns were not imported.'.format(file=file_name)
    )
    log.status.Print(repr(e))
    return [], True

  success = True
  for pattern in patterns:
    try:
      re.compile(pattern)
    except Exception as e:  # pylint: disable=broad-except
      success = False
      log.status.Print("Cannot compile regular expression '{}'".format(pattern))
      log.status.Print(repr(e))

  if not success:
    return [], True

  return patterns, False


@base.UniverseCompatible
@base.ReleaseTracks(base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA)
class Migrate(base.CreateCommand):
  """Migrate from legacy firewall rules to network firewall policies."""

  NETWORK_ARG = None
  exclusion_patterns = [
      # https://cloud.google.com/kubernetes-engine/docs/concepts/firewall-rules
      # gke-[cluster-hash]-ipv6-all
      'gke-(.+)-ipv6-all',
      # gke-[cluster-name]-[cluster-hash]-master
      # gke-[cluster-name]-[cluster-hash]-vms
      # gke-[cluster-name]-[cluster-hash]-all
      # gke-[cluster-name]-[cluster-hash]-inkubelet
      # gke-[cluster-name]-[cluster-hash]-exkubelet
      # gke-[cluster-name]-[cluster-hash]-mcsd
      'gke-(.+)-(.+)-((master)|(vms)|(all)|(inkubelet)|(exkubelet)|(mcsd))',
      # k8s-fw-[loadbalancer-hash]
      # k8s-fw-l7-[random-hash]
      'k8s-fw-(l7-)?(.+)',
      # k8s-[cluster-id]-node-http-hc
      # k8s-[loadbalancer-hash]-http-hc
      # k8s-[cluster-id]-node-hc
      'k8s-(.+)-((node)|(http)|(node-http))-hc',
      # [loadbalancer-hash]-hc
      '(.+)-hc',
      # k8s2-[cluster-id]-[namespace]-[service-name]-[suffixhash]
      # k8s2-[cluster-id]-[namespace]-[service-name]-[suffixhash]-fw
      'k8s2-(.+)-(.+)-(.+)-(.+)(-fw)?',
      # k8s2-[cluster-id]-l4-shared-hc-fw
      'k8s2-(.+)-l4-shared-hc-fw',
      # gkegw1-l7-[network]-[region/global]
      # gkemcg1-l7-[network]-[region/global]
      'gke((gw)|(mcg))1-l7-(.+)-(.+)',
      # https://cloud.google.com/vpc/docs/serverless-vpc-access#firewall_rules
  ]

  @classmethod
  def Args(cls, parser):
    # optional --target-firewall-policy=TARGET_FIREWALL_POLICY argument
    group = parser.add_group(mutex=True, required=True)
    group.add_argument(
        '--target-firewall-policy',
        help="""\
      Name of the new Network Firewall Policy used to store the migration
      result.
      """,
    )
    group.add_argument(
        '--export-tag-mapping',
        action='store_true',
        help="""\
      If set, migration tool will inspect all VPC Firewalls attached to
      SOURCE_NETWORK, collect all source and target tags, and store them in
      TAG_MAPPING_FILE.
      """,
    )
    group.add_argument(
        '--export-exclusion-patterns',
        action='store_true',
        help="""\
      If set, migration tool will dump list of regexes used to filter VPC Firewall out of migration.
      """,
    )
    group.add_argument(
        '--bind-tags-to-instances',
        action='store_true',
        help="""\
      If set, migration tool will bind secure tags to the instances with the
      network tags which match secure tags from the tag mapping file.
      """,
    )
    # required --source-network=NETWORK flag
    cls.NETWORK_ARG = compute_flags.ResourceArgument(
        name='--source-network',
        resource_name='network',
        completer=network_flags.NetworksCompleter,
        plural=False,
        required=True,
        global_collection='compute.networks',
        short_help=(
            'The VPC Network for which the migration should be performed.'
        ),
        detailed_help=None,
    )
    cls.NETWORK_ARG.AddArgument(parser)
    # optional --tag-mapping-file=TAG_MAPPING_FILE argument
    parser.add_argument(
        '--tag-mapping-file',
        required=False,
        help=(
            'Path to a JSON file with legacy tags and service accounts to'
            ' secure tags mapping.'
        ),
    )
    # optional --export-terraform-script argument
    parser.add_argument(
        '--export-terraform-script',
        action='store_true',
        required=False,
        help=(
            'If set, migration tool will output a terraform script to create a'
            ' Firewall Policy with migrated rules.'
        ),
    )
    # optional --terraform-script-output-file=TERRAFORM_SCRIPT_OUTPUT_FILE arg
    parser.add_argument(
        '--terraform-script-output-file',
        required=False,
        help='Path to a file where to store generated Terraform script.',
    )
    # optional --exclusion-patterns-file=EXCLUSION_PATTERNS_FILE argument
    parser.add_argument(
        '--exclusion-patterns-file',
        required=False,
        help=(
            'Path to a file with exclusion patterns used for VPC Firewall'
            ' filtering. Each regular expression describing a single firewall'
            ' naming pattern must be placed in a single line. No leading or'
            ' tailing whitespaces.'
        ),
    )
    # optional --force argument
    parser.add_argument(
        '--force',
        action='store_true',
        required=False,
        help=(
            'If set, migration will succeed even if the tool detects that'
            ' original rule evaluation order cannot be preserved.'
        ),
    )
    # optional --force argument
    parser.add_argument(
        '--skip-migrate-target-service-accounts-to-tags',
        action='store_true',
        required=False,
        help=(
            'If set, migration will keep target service accounts as they are'
            ' and will not try to replace them with secure tags.'
        ),
    )

  def Run(self, args):
    """Run the migration logic."""
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    client = holder.client.apitools_client
    messages = client.MESSAGES_MODULE

    # Determine project
    if args.project:
      project = args.project
    else:
      project = properties.VALUES.core.project.GetOrFail()

    # Get Input Parameters
    network_name = getattr(args, 'source_network')
    policy_name = getattr(args, 'target_firewall_policy', None)
    export_tag_mapping = getattr(args, 'export_tag_mapping', False)
    tag_mapping_file_name = getattr(args, 'tag_mapping_file', None)
    export_exclusion_patterns = getattr(
        args, 'export_exclusion_patterns', False
    )
    exclusion_patterns_file_name = getattr(
        args, 'exclusion_patterns_file', None
    )
    bind_tags_to_instances = getattr(args, 'bind_tags_to_instances', False)
    export_terraform_script = getattr(args, 'export_terraform_script', False)
    terraform_script_output_file_name = getattr(
        args, 'terraform_script_output_file', None
    )
    force = getattr(args, 'force', False)
    keep_target_service_accounts = getattr(
        args, 'skip_migrate_target_service_accounts_to_tags', False
    )

    # In the export tag mode, the tag mapping file must be provided
    if export_tag_mapping and not tag_mapping_file_name:
      raise exceptions.ToolException(
          '--tag-mapping-file must be specified if --export-tag-mapping is set.'
      )

    # In the export patterns mode, the output file must be provided
    if export_exclusion_patterns and not exclusion_patterns_file_name:
      raise exceptions.ToolException(
          '--exclusion-patterns-file must be specified if'
          ' --export-exclusion-patterns is set.'
      )

    if bind_tags_to_instances and not tag_mapping_file_name:
      raise exceptions.ToolException(
          '--tag-mapping-file must be specified if --bind-tags-to-instances is'
          ' set.'
      )

    # Branch 0a: Bind Secure Tags to Instances
    if bind_tags_to_instances:
      _BindSecureTagsToInstances(
          network_name, project, tag_mapping_file_name, client
      )
      return

    # Branch 0b: Dump exclusion patterns used for filtering
    if export_exclusion_patterns:
      _WriteExclusionPatterns(
          exclusion_patterns_file_name, self.exclusion_patterns
      )
      log.status.Print(
          "Exclusion patterns were exported to '{}'".format(
              exclusion_patterns_file_name
          )
      )
      return

    # Import externally defined exclusion patterns if defined
    if exclusion_patterns_file_name:
      patterns, err = _ReadExclusionPatterns(exclusion_patterns_file_name)
      self.exclusion_patterns = patterns
      if err:
        log.status.Print(
            'Could not import exclusion patterns. Migration cannot be'
            ' completed.'
        )
        return

    # Get VPC Network
    network = client.networks.Get(
        messages.ComputeNetworksGetRequest(
            project=project, network=network_name
        )
    )

    log.status.Print(
        'Looking for VPC Firewalls and Network Firewall Policies associated'
        " with VPC Network '{}'.".format(network_name)
    )

    # Get all Firewall Policies
    fp_list_response = client.networkFirewallPolicies.List(
        messages.ComputeNetworkFirewallPoliciesListRequest(project=project)
    )

    # Verify there is no Firewall Policy with provided name
    for firewall_policy in fp_list_response.items:
      if firewall_policy.name == policy_name:
        log.status.Print(
            'Firewall Policy "' + policy_name + '" already exists.'
        )
        return

    # Filter Network Firewall Policies attached to the given VPC Network
    firewall_policies = _GetFirewallPoliciesAssociatedWithNetwork(
        network, fp_list_response.items
    )
    log.status.Print(
        'Found {} Network Firewall Policies associated with the VPC Network'
        " '{}'.".format(len(firewall_policies), network_name)
    )

    # Migration tool does not support multiple FirewallPolicies.
    if len(firewall_policies) > 1:
      log.status.Print(
          'Migration tool does not support multiple Network Firewall Policies '
          'associated with a single VPC Network.'
      )
      return

    # List all legacy VPC Firewalls attached to the given VPC Network
    # Hidden VPC Firewalls are not listed.
    fetched_firewalls = list_pager.YieldFromList(
        service=client.firewalls,
        batch_size=500,
        request=messages.ComputeFirewallsListRequest(project=project),
        method='List',
        field='items',
    )
    associated_firewalls = _GetFirewallsAssociatedWithNetwork(
        network, fetched_firewalls
    )
    log.status.Print(
        "Found {} VPC Firewalls associated with the VPC Network '{}'.\n".format(
            len(associated_firewalls), network_name
        )
    )

    # Now we fetched all VPC Firewalls and Firewall Policies attached to the
    # given VPC Network.

    # Filter VPC Firewalls first
    # Add unique ID to each firewall - required for sorting
    log.status.Print(
        '{} pattern(s) used to filter VPC Firewalls out:'.format(
            len(self.exclusion_patterns)
        )
    )
    for pattern in self.exclusion_patterns:
      log.status.Print(pattern)
    log.status.Print('\n')

    firewall_id = 0
    marked_firewalls = []
    for firewall in associated_firewalls:
      selected = not _IsExcludedFirewall(firewall, self.exclusion_patterns)
      marked_firewalls.append((firewall, selected, firewall_id))
      firewall_id = firewall_id + 1

    # Branch 1: Just generate pre-mapping for legacy tags
    if export_tag_mapping:
      selected_firewalls = []
      for firewall, selected, _ in marked_firewalls:
        if selected:
          selected_firewalls.append(firewall)
      legacy_tags = _GetLegacyTags(selected_firewalls)
      service_accounts = _GetServiceAccounts(
          selected_firewalls, keep_target_service_accounts
      )
      _WriteTagMapping(tag_mapping_file_name, legacy_tags, service_accounts)
      log.status.Print(
          "Legacy tags were exported to '{}'".format(tag_mapping_file_name)
      )
      return

    # Branch 2: Do the actual migration

    # Read tag mapping if provided
    tag_mapping = dict()
    if tag_mapping_file_name:
      tag_mapping = _ReadTagMapping(tag_mapping_file_name)
      if tag_mapping is None:
        log.status.Print('Stop processing, missing tag mapping file...')
        return

    # Sort VPC Firewalls by priorities. If two Firewalls have the same priority
    # then deny rules should precede allow rules. Third coordinate is unique to
    # avoid comparison between Firewall objects which is undefined.
    sorted_firewalls = [
        (f.priority, 0 if f.denied else 1, id, f, selected)
        for (f, selected, id) in marked_firewalls
    ]
    sorted_firewalls = sorted(sorted_firewalls)

    # Convert user provided VPC Firewalls if possible
    converted_firewalls = []
    conversion_failures_count = 0
    selected_firewalls_count = 0
    for priority, _, _, firewall, selected in sorted_firewalls:
      (status, error) = (False, 'Not a customer defined VPC Firewall.')
      converted_firewall = None
      # Convert only supported selected VPC Firewalls
      if selected:
        selected_firewalls_count = selected_firewalls_count + 1
        (status, error) = _IsFirewallSupported(
            firewall, tag_mapping, keep_target_service_accounts
        )
        if status:
          converted_firewall = _ConvertRule(
              messages,
              firewall,
              tag_mapping,
              keep_target_service_accounts,
          )
        else:
          conversion_failures_count = conversion_failures_count + 1
      converted_firewalls.append(
          (priority, firewall, selected, converted_firewall, status, error)
      )

    # Print info about VPC Firewalls [both selected and unselected].
    if selected_firewalls_count:
      log.status.Print(
          'Found {} selected VPC Firewalls.'.format(selected_firewalls_count)
      )
      log.status.Print("priority: name 'description'")
      for _, firewall, selected, _, _, _ in converted_firewalls:
        if selected:
          log.status.Print(
              "{}: {} '{}'".format(
                  firewall.priority, firewall.name, firewall.description
              )
          )
      log.status.Print('')

    non_selected_firewall_count = (
        len(converted_firewalls) - selected_firewalls_count
    )
    if non_selected_firewall_count:
      log.status.Print(
          '{} VPC Firewalls were not selected.'.format(
              non_selected_firewall_count
          )
      )
      log.status.Print("priority: name 'description'")
      for _, firewall, selected, _, _, _ in converted_firewalls:
        if not selected:
          log.status.Print(
              "{}: {} '{}'".format(
                  firewall.priority, firewall.name, firewall.description
              )
          )
      log.status.Print('')

    # Print info about conversion failures
    if conversion_failures_count:
      log.status.Print(
          'Could not convert {} selected VPC Firewalls:'.format(
              conversion_failures_count
          )
      )
      for _, firewall, _, _, status, error in converted_firewalls:
        if not status:
          log.status.Print(
              '{}: {} - {}'.format(firewall.priority, firewall.name, error)
          )
      log.status.Print('')

    # Filter out default FirewallPolicy.Rules
    # There is at most one firewall policy to iterate on.
    firewall_policy_rules = []
    for firewall_policy in firewall_policies:
      for rule in firewall_policy.rules:
        if not _IsDefaultFirewallPolicyRule(rule):
          firewall_policy_rules.append(rule)

    # Sort FirewallPolicy.Rules by priority
    firewall_policy_rules = [
        (rule.priority, rule) for rule in firewall_policy_rules
    ]
    firewall_policy_rules = sorted(firewall_policy_rules)

    # Adjust the format to match list of converted VPC Firewalls
    firewall_policy_rules = [
        # (priority, vpc_firewall_rule, selected, converted_rule, status, error)
        (priority, None, True, rule, True, '')
        for (priority, rule) in firewall_policy_rules
    ]

    # Join converted selected VPC Firewalls with Network Firewall Policy Rules
    joined_rules = []
    if (
        network.networkFirewallPolicyEnforcementOrder
        == messages.Network.NetworkFirewallPolicyEnforcementOrderValueValuesEnum.AFTER_CLASSIC_FIREWALL
    ):
      joined_rules.extend(converted_firewalls)
      joined_rules.extend(firewall_policy_rules)
    else:
      joined_rules.extend(firewall_policy_rules)
      joined_rules.extend(converted_firewalls)

    # Important: Non-selected vpc firewalls are still present on the list!
    # However, they are marked as non-selected.

    # Check if extraction of selected rules is possible
    # Extracted rules must be a prefix for BEFORE_CLASSIC_FIREWALL mode and
    # suffix for AFTER_CLASSIC_FIREWALL mode.
    statuses = [status for (_, _, _, _, status, _) in joined_rules]
    safe_migration_impossible = False
    safe_migration_error_message = (
        'Safe migration is impossible, because rule evaluation order cannot be '
        'preserved.'
    )
    if (
        network.networkFirewallPolicyEnforcementOrder
        == messages.Network.NetworkFirewallPolicyEnforcementOrderValueValuesEnum.AFTER_CLASSIC_FIREWALL
    ):
      if not _IsSuffixTrue(statuses):
        log.status.Print(safe_migration_error_message)
        safe_migration_impossible = True
    else:
      if not _IsPrefixTrue(statuses):
        log.status.Print(safe_migration_error_message)
        safe_migration_impossible = True

    # Stop migration if not possible to do it safely
    if safe_migration_impossible:
      if force:
        log.status.Print(
            'WARNING: Forcing migration of chosen firewall rules.\n'
        )
      else:
        return

    # Extract rules to migrate
    rules_to_migrate = [(p, r, f) for (p, f, _, r, s, _) in joined_rules if s]

    # Check if priorities remap is needed
    priorities_remap_needed = len(
        set([p for (p, r, f) in rules_to_migrate])
    ) != len(rules_to_migrate)

    # Compute new priorities if needed
    if priorities_remap_needed:
      log.status.Print('Updating rules priorities to avoid collisions.')
      log.status.Print(
          "new-priority: old-priority rule-name 'rule-description'"
      )

    current_priority = 1000
    migrated_rules = []
    for priority, rule, firewall in rules_to_migrate:
      if priorities_remap_needed:
        rule.priority = current_priority
        current_priority = current_priority + 1
        log.status.Print(
            "{}: {} {} '{}'".format(
                rule.priority, priority, rule.ruleName, rule.description
            )
        )
      migrated_rules.append((rule, firewall))
    if priorities_remap_needed:
      log.status.Print('')

    # Create a new Network Firewall Policy
    if self.ReleaseTrack() == base.ReleaseTrack.ALPHA:
      firewall_policy = messages.FirewallPolicy(
          description=(
              'Network Firewall Policy containing all VPC Firewalls and'
              ' FirewallPolicy.Rules migrated using GCP Firewall Migration'
              ' Tool.'
          ),
          name=policy_name,
          vpcNetworkScope=messages.FirewallPolicy.VpcNetworkScopeValueValuesEnum.GLOBAL_VPC_NETWORK,
      )
    else:
      firewall_policy = messages.FirewallPolicy(
          description=(
              'Network Firewall Policy containing all VPC Firewalls and'
              ' FirewallPolicy.Rules migrated using GCP Firewall Migration'
              ' Tool.'
          ),
          name=policy_name,
      )

    if export_terraform_script:
      tf_script = (
          convert_terraform.ConvertFirewallPolicy(firewall_policy, project)
          + '\n'
      )
      for rule, _ in migrated_rules:
        tf_script = (
            tf_script + convert_terraform.ConvertFirewallPolicyRule(rule) + '\n'
        )
      if terraform_script_output_file_name:
        _WriteTerraformScript(terraform_script_output_file_name, tf_script)
        log.status.Print(
            "Terraform script exported to the file '{}'".format(
                terraform_script_output_file_name
            )
        )
      else:
        log.status.Print(
            'Terraform script for migrated Network Firewall Policy:'
        )
        log.status.Print(tf_script)
      return

    response = client.networkFirewallPolicies.Insert(
        messages.ComputeNetworkFirewallPoliciesInsertRequest(
            project=project, firewallPolicy=firewall_policy
        )
    )
    # Wait until Network Firewall Policy is created
    operation_poller = poller.Poller(client.networkFirewallPolicies)
    operation_ref = holder.resources.Parse(
        response.selfLink, collection='compute.globalOperations'
    )
    waiter.WaitFor(
        operation_poller,
        operation_ref,
        "Creating new Network Firewall Policy '{}'".format(policy_name),
    )

    # Add migrated rules to newly created policy
    log.status.Print('Migrating the following VPC Firewalls:')
    log.status.Print("old-priority: rule-name 'rule-description'")
    responses = []
    for rule, firewall in migrated_rules:
      responses.append(
          client.networkFirewallPolicies.AddRule(
              messages.ComputeNetworkFirewallPoliciesAddRuleRequest(
                  firewallPolicy=policy_name,
                  firewallPolicyRule=rule,
                  project=project,
              )
          )
      )
      if firewall:
        log.status.Print(
            "{}: {} '{}'".format(
                firewall.priority, firewall.name, firewall.description
            )
        )
    # Wait until rules are added
    operation_poller = poller.BatchPoller(
        holder.client, client.networkFirewallPolicies
    )
    operation_refs = [
        holder.resources.Parse(
            response.selfLink, collection='compute.globalOperations'
        )
        for response in responses
    ]
    waiter.WaitFor(
        operation_poller, poller.OperationBatch(operation_refs), 'Migrating'
    )


Migrate.detailed_help = {
    'brief': (
        'Create a new Network Firewall Policy and move all customer defined '
        'firewall rules there.'
    ),
    'DESCRIPTION': """
*{command}* is used to create a new Network Firewall Policy that contain
all rules defined in already existing Network Firewall Policy associated with
the given VPC Network and all customer defined VPC Firewall Rules attached to
that VPC Network.
""",
    'EXAMPLES': """
To execute the migration for VPC Network 'my-network' which stores the result
in 'my-policy' Network Firewall Policy, run:

  $ {command} --source-network=my-network --target-firewall-policy=my-policy
  """,
}



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""API wrapper for `gcloud service-extensions wasm-plugin` commands."""


from googlecloudsdk.api_lib.service_extensions import util
from googlecloudsdk.api_lib.util import waiter
from googlecloudsdk.calliope import base
from googlecloudsdk.core import resources


class Client:
  """API client for WasmPlugin commands.

  Attributes:
    messages: API messages class, The Networkservices API messages.
  """

  def __init__(self, release_track):
    self._client = util.GetClientInstance(release_track)
    self._wasm_plugin_client = self._client.projects_locations_wasmPlugins
    self._operations_client = self._client.projects_locations_operations
    self.messages = self._client.MESSAGES_MODULE
    self._resource_parser = resources.Registry()
    self._resource_parser.RegisterApiByName(
        'networkservices', util.API_VERSION_FOR_TRACK.get(release_track)
    )
    self._release_track = release_track

  def CreateWasmPlugin(
      self,
      name,
      parent,
      description=None,
      labels=None,
      log_config=None,
      main_version=None,
      versions=None,
      kms_key_name=None,
  ):
    """Calls the CreateWasmPlugin API."""
    if self._release_track in [base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA]:
      wasm_plugin = self.messages.WasmPlugin(
          description=description,
          labels=labels,
          logConfig=log_config,
          mainVersionId=main_version,
          versions=versions,
          kmsKeyName=kms_key_name,
      )
    else:
      wasm_plugin = self.messages.WasmPlugin(
          description=description,
          labels=labels,
          logConfig=log_config,
          mainVersionId=main_version,
          versions=versions,
      )
    request = (
        self.messages.NetworkservicesProjectsLocationsWasmPluginsCreateRequest(
            parent=parent,
            wasmPluginId=name,
            wasmPlugin=wasm_plugin,
        )
    )
    return self._wasm_plugin_client.Create(request)

  def UpdateWasmPlugin(
      self,
      name,
      main_version,
      update_mask=None,
      description=None,
      labels=None,
      log_config=None,
      kms_key_name=None,
  ):
    """Calls the UpdateWasmPlugin API."""
    if self._release_track in [base.ReleaseTrack.ALPHA, base.ReleaseTrack.BETA]:
      wasm_plugin = self.messages.WasmPlugin(
          mainVersionId=main_version,
          description=description,
          labels=labels,
          logConfig=log_config,
          kmsKeyName=kms_key_name,
      )
    else:
      wasm_plugin = self.messages.WasmPlugin(
          mainVersionId=main_version,
          description=description,
          labels=labels,
          logConfig=log_config,
      )
    request = (
        self.messages.NetworkservicesProjectsLocationsWasmPluginsPatchRequest(
            name=name,
            updateMask=update_mask,
            wasmPlugin=wasm_plugin,
        )
    )
    return self._wasm_plugin_client.Patch(request)

  def PrepareVersionDetailsForSingleVersion(
      self, version_id, image, plugin_config_data, plugin_config_uri
  ):
    if version_id is None:
      return None
    return self.messages.WasmPlugin.VersionsValue(
        additionalProperties=[
            self.messages.WasmPlugin.VersionsValue.AdditionalProperty(
                key=version_id,
                value=self.messages.WasmPluginVersionDetails(
                    imageUri=image,
                    pluginConfigData=plugin_config_data,
                    pluginConfigUri=plugin_config_uri,
                ),
            )
        ]
    )

  def WaitForOperation(
      self,
      operation_ref,
      message,
  ):
    """Waits for the opration to complete and returns the result of the operation.

    Args:
      operation_ref: A Resource describing the Operation.
      message: The message to display to the user while they wait.

    Returns:
      result of result_service.Get request for the provided operation.
    """

    op_resource = resources.REGISTRY.ParseRelativeName(
        operation_ref.name,
        collection='networkservices.projects.locations.operations',
    )
    poller = waiter.CloudOperationPoller(
        self._wasm_plugin_client, self._operations_client
    )
    return waiter.WaitFor(
        poller,
        op_resource,
        message,
    )



# -*- coding: utf-8 -*- #
# Copyright 2014 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command for describing backend services."""


from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.compute.backend_services import backend_services_utils
from googlecloudsdk.command_lib.compute.backend_services import flags


class Describe(base.DescribeCommand):
  """Describe a backend service."""

  @staticmethod
  def Args(parser):
    flags.GLOBAL_REGIONAL_BACKEND_SERVICE_ARG.AddArgument(
        parser, operation_type='describe')

  def Run(self, args):
    """Issues request necessary to describe the backend service."""
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    client = holder.client

    (backend_services_utils.
     IsDefaultRegionalBackendServicePropertyNoneWarnOtherwise())
    backend_service_ref = (
        flags.GLOBAL_REGIONAL_BACKEND_SERVICE_ARG.ResolveAsResource(
            args,
            holder.resources,
            scope_lister=compute_flags.GetDefaultScopeLister(client)))

    if backend_service_ref.Collection() == 'compute.backendServices':
      service = client.apitools_client.backendServices
      request = client.messages.ComputeBackendServicesGetRequest(
          **backend_service_ref.AsDict())
    elif backend_service_ref.Collection() == 'compute.regionBackendServices':
      service = client.apitools_client.regionBackendServices
      request = client.messages.ComputeRegionBackendServicesGetRequest(
          **backend_service_ref.AsDict())

    return client.MakeRequests([(service, 'Get', request)])[0]


Describe.detailed_help = base_classes.GetMultiScopeDescriberHelp(
    'backend service', [base_classes.ScopeType.regional_scope,
                        base_classes.ScopeType.global_scope])



# -*- coding: utf-8 -*- #
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to list NATs on a Compute Engine router."""


import textwrap

from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.compute.routers import flags as routers_flags
from googlecloudsdk.command_lib.compute.routers.nats import nats_utils
from googlecloudsdk.command_lib.compute.routers.nats.rules import flags as rules_flags


class List(base.DescribeCommand):
  """Lists the NATs on a Compute Engine router."""

  @classmethod
  def Args(cls, parser):
    cls.ROUTER_ARG = routers_flags.RouterArgumentForNat()
    cls.ROUTER_ARG.AddArgument(parser)

    rules_flags.AddNatNameArg(parser)

    parser.display_info.AddFormat(rules_flags.DEFAULT_LIST_FORMAT)

    compute_flags.AddRegionFlag(
        parser, 'NAT containing the Rules', operation_type='list')

  def Run(self, args):
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    messages = holder.client.messages
    service = holder.client.apitools_client.routers

    router_ref = self.ROUTER_ARG.ResolveAsResource(args, holder.resources)

    request_type = messages.ComputeRoutersGetRequest
    router = service.Get(request_type(**router_ref.AsDict()))

    nat_name = args.nat

    nat = nats_utils.FindNatOrRaise(router, nat_name)

    return nat.rules


List.detailed_help = {
    'DESCRIPTION':
        textwrap.dedent("""\
        *{command}* is used to list the Rule on a Compute Engine NAT.
     """),
    'EXAMPLES':
        """\
    To list all Rules in Nat ``n1'' in router ``r1'' in region ``us-central1'',
    run:

        $ {command} --nat=n1 --router=r1 --region=us-central1.
    """
}



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for getting NAT IP information from Compute Engine routers."""


from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.compute.routers import flags as routers_flags


class GetNatIpInfo(base.DescribeCommand):
  """Display NAT IP information in a router."""

  ROUTER_ARG = None

  @classmethod
  def Args(cls, parser):
    cls.ROUTER_ARG = routers_flags.RouterArgument()
    cls.ROUTER_ARG.AddArgument(parser, operation_type='get NAT IP info')
    routers_flags.AddGetNatIpInfoArgs(parser)
    base.URI_FLAG.RemoveFromParser(parser)

  def Run(self, args):
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    client = holder.client

    router_ref = self.ROUTER_ARG.ResolveAsResource(
        args,
        holder.resources,
        scope_lister=compute_flags.GetDefaultScopeLister(client),
    )

    params = router_ref.AsDict()
    if args.nat_name:
      params['natName'] = args.nat_name

    request = client.messages.ComputeRoutersGetNatIpInfoRequest(**params)

    return client.MakeRequests([(client.apitools_client.routers,
                                 'GetNatIpInfo', request)])[0]

GetNatIpInfo.detailed_help = {
    'DESCRIPTION': """
        $ {command}

        shows a mapping of IP:[usage, mode]
        allocated to each NAT via the specified router.""",
    'EXAMPLES': """\
        To show NAT IP information from all NATs in router 'r1' in region
        'us-central1', run:

            $ {command} r1 --region=us-central1

        To show NAT IP information for a specific NAT 'nat1' in router 'r1' in
        region 'us-central1', run:

            $ {command} r1 --region=us-central1 --nat-name="nat1"
        """
}



# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General utilities for copies."""


import datetime
import os

from googlecloudsdk.api_lib.storage import cloud_api
from googlecloudsdk.api_lib.storage import errors as api_errors
from googlecloudsdk.command_lib.storage import errors as command_errors
from googlecloudsdk.command_lib.storage import manifest_util
from googlecloudsdk.command_lib.storage import storage_url
from googlecloudsdk.command_lib.storage.resources import resource_reference
from googlecloudsdk.command_lib.storage.resources import resource_util
from googlecloudsdk.command_lib.storage.tasks import task
from googlecloudsdk.core import exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import properties

_EARLY_DELETION_MINIMUM_DAYS = {
    'nearline': 30,
    'coldline': 90,
    'archive': 365,
}


class CopyTask(task.Task):
  """Parent task that handles common attributes and an __init__ status print."""

  def __init__(
      self,
      source_resource,
      destination_resource,
      print_created_message=False,
      print_source_version=False,
      user_request_args=None,
      verbose=False,
  ):
    """Initializes task.

    Args:
      source_resource (resource_reference.Resource): Source resource to copy.
      destination_resource (resource_reference.Resource): Target resource to
        copy to.
      print_created_message (bool): Print a message containing the URL of the
        copy result.
      print_source_version (bool): Print source object version in status message
        enabled by the `verbose` kwarg.
      user_request_args (UserRequestArgs|None): Various user-set values
        typically converted to an API-specific RequestConfig.
      verbose (bool): Print a "copying" status message on initialization.
    """
    super(CopyTask, self).__init__()
    self._source_resource = source_resource
    self._destination_resource = destination_resource
    self._print_created_message = print_created_message
    self._print_source_version = print_source_version
    self._user_request_args = user_request_args
    self._verbose = verbose

    self._send_manifest_messages = bool(
        self._user_request_args and self._user_request_args.manifest_path
    )

    if verbose:
      if self._print_source_version:
        source_string = source_resource.storage_url.url_string
      else:
        source_string = source_resource.storage_url.versionless_url_string
      log.status.Print(
          'Copying {} to {}'.format(
              source_string,
              destination_resource.storage_url.versionless_url_string,
          )
      )

  def _print_created_message_if_requested(self, resource):
    if self._print_created_message:
      log.status.Print('Created: {}'.format(resource))


class ObjectCopyTask(CopyTask):
  """Parent task that handles common attributes for object copy tasks."""

  def __init__(
      self,
      source_resource,
      destination_resource,
      posix_to_set=None,
      print_created_message=False,
      print_source_version=False,
      user_request_args=None,
      verbose=False,
  ):
    """Initializes task.

    Args:
      source_resource (resource_reference.Resource): See parent class.
      destination_resource (resource_reference.Resource): See parent class.
      posix_to_set (PosixAttributes|None): POSIX info set as custom cloud
        metadata on target.
      print_created_message (bool): See parent class.
      print_source_version (bool): See parent class.
      user_request_args (UserRequestArgs|None): See parent class.
      verbose (bool): Print a "copying" status message on initialization.
    """
    self._posix_to_set = posix_to_set
    # Set before super().__init__ call because otherwise the attribute won't be
    # available for the _get_source_string_for_status_message call.
    self._print_source_version = print_source_version

    super(ObjectCopyTask, self).__init__(
        source_resource,
        destination_resource,
        print_created_message,
        print_source_version,
        user_request_args,
        verbose,
    )


class _ExitHandlerMixin:
  """Provides an exit handler for copy tasks."""

  def exit_handler(self, error=None, task_status_queue=None):
    """Send copy result info to manifest if requested."""
    if error and self._send_manifest_messages:
      if not task_status_queue:
        raise command_errors.Error(
            'Unable to send message to manifest for source: {}'.format(
                self._source_resource
            )
        )
      manifest_util.send_error_message(task_status_queue, self._source_resource,
                                       self._destination_resource, error)


class CopyTaskWithExitHandler(
    # _ExitHandlerMixin must precede CopyTask, otherwise task.Task.exit_hander
    # overrides the intended implementation.
    _ExitHandlerMixin,
    CopyTask,
):
  """Parent task with an exit handler for non-object copy tasks."""


class ObjectCopyTaskWithExitHandler(_ExitHandlerMixin, ObjectCopyTask):
  """Parent task with an exit handler for object copy tasks."""


def get_no_clobber_message(destination_url):
  """Returns standardized no clobber warning."""
  return 'Skipping existing destination item (no-clobber): {}'.format(
      destination_url)


def check_for_cloud_clobber(user_request_args, api_client,
                            destination_resource):
  """Returns if cloud destination object exists if no-clobber enabled."""
  if not (user_request_args and user_request_args.no_clobber):
    return False
  try:
    api_client.get_object_metadata(
        destination_resource.storage_url.bucket_name,
        destination_resource.storage_url.resource_name,
        fields_scope=cloud_api.FieldsScope.SHORT)
  except api_errors.NotFoundError:
    return False
  return True


def get_generation_match_value(request_config):
  """Prioritizes user-input generation over no-clobber zero value."""
  if request_config.precondition_generation_match is not None:
    return request_config.precondition_generation_match
  if request_config.no_clobber:
    return 0
  return None


def raise_if_mv_early_deletion_fee_applies(object_resource):
  """Raises error if Google Cloud Storage object will incur an extra charge."""
  if isinstance(object_resource, resource_reference.FolderResource):
    return
  if not (properties.VALUES.storage.check_mv_early_deletion_fee.GetBool() and
          object_resource.storage_url.scheme is storage_url.ProviderPrefix.GCS
          and object_resource.creation_time and
          object_resource.storage_class in _EARLY_DELETION_MINIMUM_DAYS):
    return

  minimum_lifetime = _EARLY_DELETION_MINIMUM_DAYS[
      object_resource.storage_class.lower()]
  creation_datetime_utc = resource_util.convert_datetime_object_to_utc(
      object_resource.creation_time)
  current_datetime_utc = resource_util.convert_datetime_object_to_utc(
      datetime.datetime.now())

  if current_datetime_utc < creation_datetime_utc + datetime.timedelta(
      days=minimum_lifetime):
    raise exceptions.Error(
        ('Deleting {} may incur an early deletion charge. Note: the source'
         ' object of a mv operation is deleted.\nThe object appears to have'
         ' been created on {}, and the minimum time before deletion for the {}'
         ' storage class is {} days.\nTo allow deleting the object anyways, run'
         ' "gcloud config set storage/check_mv_early_deletion_fee False"'
        ).format(object_resource, object_resource.creation_time,
                 object_resource.storage_class, minimum_lifetime))


def should_skip_file_download_if_outside_destination(
    source_resource_url,
    source_expanded_url,
    destination_url,
    raw_destination_url,
):
  """Returns True if the file downloads outside the destination directory."""

  # Traversal Detection:
  # POSIX-safe check: '\' is a valid filename character on Linux/macOS,
  # so we only evaluate '..\' as a traversal on Windows.
  source_key = source_resource_url.resource_name
  has_traversal = ('../' in source_key
                   or (os.sep == '\\' and '..\\' in source_key))

  # If the source object key contains no traversal sequences, it's safe.
  # This fast-path avoids regressions in recursive copies of safe files
  # on cross-platform mocked tests.
  if not has_traversal:
    return False

  # If this is not a recursive download to a file, then we can skip this check.
  if not (
      isinstance(source_resource_url, storage_url.CloudUrl)
      and isinstance(destination_url, storage_url.FileUrl)
      and isinstance(raw_destination_url, storage_url.FileUrl)
      # a resource_name check prevents abspath('') from defaulting to CWD.
      and raw_destination_url.resource_name
  ):
    return False

  # If downloading into a directory, keys starting with '..' are inherently
  # escape attempts. We must check this before the bounds check because the CLI
  # automatically strips leading traversals from the final local destination
  # path, which makes them look deceptively "safe" to the bounds check below.
  is_recursive_download = (
      source_resource_url.versionless_url_string
      != source_expanded_url.versionless_url_string
  )
  is_directory_download = (
      is_recursive_download
      or raw_destination_url.isdir()
      or raw_destination_url.resource_name.endswith(os.sep)
  )

  if is_directory_download:
    stripped_key = source_key.lstrip('/')
    key_for_start_check = (stripped_key.replace('\\', '/')
                           if os.sep == '\\' else stripped_key)
    if key_for_start_check.startswith('../'):
      return True

  container = os.path.abspath(
      os.path.normpath(raw_destination_url.resource_name)
  )
  destination_path = os.path.abspath(
      os.path.normpath(destination_url.resource_name)
  )

  # For single-file explicit renames (e.g. `cp gs://b/../a.txt ./b.txt`),
  # landing exactly on the container path is intended behavior.
  if not is_directory_download and destination_path == container:
    return False

  # In a recursive copy, files are placed inside the destination directory,
  # so we don't use an exact path match. `startswith` is used instead, but
  # it must be made safe against matching similar paths (e.g. /a/b
  # matching /a/bc) by ensuring a path separator is present.
  if destination_path.startswith(os.path.join(container, '')):
    return False

  return True



# -*- coding: utf-8 -*- #
# Copyright 2023 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command to detach a Service Project."""

from googlecloudsdk.api_lib.apphub import service_projects as apis
from googlecloudsdk.calliope import base
from googlecloudsdk.core import properties

_DETAILED_HELP = {
    'DESCRIPTION': '{description}',
    'EXAMPLES': """ \
        To detach the service project `my-service-project`, run:

          $ {command} --project=my-service-project
        """,
}


@base.ReleaseTracks(base.ReleaseTrack.GA)
class DetachGA(base.SilentCommand):
  """Detach an Apphub service project."""

  detailed_help = _DETAILED_HELP

  def Run(self, args):
    """Run the detach command."""
    client = apis.ServiceProjectsClient(release_track=base.ReleaseTrack.GA)
    service_project = properties.VALUES.core.project.Get()
    return client.Detach(service_project=service_project)


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class DetachAlpha(base.SilentCommand):
  """Detach an Apphub service project."""

  detailed_help = _DETAILED_HELP

  def Run(self, args):
    """Run the detach command."""
    client = apis.ServiceProjectsClient(release_track=base.ReleaseTrack.ALPHA)
    service_project = properties.VALUES.core.project.Get()
    return client.Detach(service_project=service_project)
